module histogram_routines
  use types
  use consts_dp

  implicit none

  type histo_resummed_type
  character(100) :: name
  character(10) :: obs
  integer       :: binvar   ! this index identifies the variation in the binning in case of runs with multiple binnings
  integer       :: scale ! this index identifies the scale in case of runs with multiple scales
  real(dp), allocatable :: right(:)
  real(dp), allocatable :: weight(:)
  real(dp), allocatable :: error(:)
  end type histo_resummed_type

  type histo_expanded_type
  character(100) :: name
  character(10) :: obs
  integer       :: binvar   ! this index identifies the variation in the binning in case of runs with multiple binnings
  integer       :: scale
  real(dp), allocatable :: lo(:)
  real(dp), allocatable :: nlo(:)
  real(dp), allocatable :: nnlo(:)
  real(dp), allocatable :: n3lo(:)
  end type histo_expanded_type

  type histo_lumi_type
  character(100) :: name
  character(10) :: obs
  integer       :: binvar   ! this index identifies the variation in the binning in case of runs with multiple binnings
  integer       :: scale
  real(dp)  :: asympt
  real(dp)  :: as
  real(dp)  :: as2
  real(dp)  :: as3
  end type histo_lumi_type


  type histo_resum
  type(histo_resummed_type) :: values
  end type histo_resum

  type histo_expansion
  type(histo_expanded_type) :: values
  end type histo_expansion

  type luminosity
  type(histo_lumi_type) :: values
  end type luminosity

  !---------------------------------------------------------------------

  public :: book_histogram, fill_histogram, fill_histogram_smooth
  public :: fill_histogram_array, fill_histogram_array_smooth
  public :: fill_expansion, fill_luminosity, normalise_histograms, deallocate_histograms, set_scale_index, set_binvar_index

  integer, public, save :: number_of_histograms = 0, scale_index = 1, binvar_index = 1
  logical, save :: first_call = .true.

  type(histo_resum), public, allocatable     :: histo_resummed(:)
  type(histo_expansion), public, allocatable :: histo_expanded(:)
  type(luminosity), public, allocatable      :: histo_luminosity(:)

  contains

  subroutine set_scale_index(index)
    use iso_c_binding
    integer, intent(in) :: index
    scale_index = index
    return
  end subroutine set_scale_index

  subroutine set_binvar_index(index)
    use iso_c_binding
    integer, intent(in) :: index
    binvar_index = index
    return
  end subroutine set_binvar_index

  ! add a histogram to the existing structure
  subroutine book_histogram(name, obs, scale, bins, binvar_in)
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: obs
    integer, intent(in), optional   :: binvar_in
    integer, intent(in)          :: scale
    integer :: binvar
    real(dp), intent(in)         :: bins(:) ! the array contains the bins [x0,x1,x2, ..., xN]
    type(histo_resum), allocatable     :: tmp_histo_resummed(:)
    type(histo_expansion), allocatable :: tmp_histo_expanded(:)
    type(luminosity), allocatable      :: tmp_histo_luminosity(:)
    integer :: i

    number_of_histograms = number_of_histograms + 1

    if (first_call) then
      allocate(histo_resummed(number_of_histograms), histo_expanded(number_of_histograms), histo_luminosity(number_of_histograms))
      first_call = .false.
    else
      ! copy all the content to the new histograms
      allocate(tmp_histo_resummed(number_of_histograms), tmp_histo_expanded(number_of_histograms), tmp_histo_luminosity(number_of_histograms))
      tmp_histo_resummed(:number_of_histograms - 1) = histo_resummed(:)
      tmp_histo_expanded(:number_of_histograms - 1) = histo_expanded(:)
      tmp_histo_luminosity(:number_of_histograms - 1) = histo_luminosity(:)

      ! reallocate the histograms arrays
      deallocate(histo_resummed, histo_expanded, histo_luminosity)
      allocate(histo_resummed(number_of_histograms), histo_expanded(number_of_histograms), histo_luminosity(number_of_histograms))
      histo_resummed(:)   = tmp_histo_resummed(:)
      histo_expanded(:)   = tmp_histo_expanded(:)
      histo_luminosity(:) = tmp_histo_luminosity(:)
      deallocate(tmp_histo_resummed, tmp_histo_expanded, tmp_histo_luminosity)
    end if

    ! now book the actual histogram
    allocate(histo_resummed(number_of_histograms)%values%right(size(bins)), histo_resummed(number_of_histograms)%values%weight(size(bins)),&
      &      histo_resummed(number_of_histograms)%values%error(size(bins)))
    allocate(histo_expanded(number_of_histograms)%values%lo(size(bins)), histo_expanded(number_of_histograms)%values%nlo(size(bins)),&
      & histo_expanded(number_of_histograms)%values%nnlo(size(bins)), histo_expanded(number_of_histograms)%values%n3lo(size(bins)))

    binvar=1
    if( present(binvar_in)) binvar=binvar_in

    histo_resummed(number_of_histograms)%values%obs        = obs
    histo_expanded(number_of_histograms)%values%obs        = obs
    histo_luminosity(number_of_histograms)%values%obs      = obs
    histo_resummed(number_of_histograms)%values%name       = name
    histo_expanded(number_of_histograms)%values%name       = name
    histo_luminosity(number_of_histograms)%values%name     = name
    histo_resummed(number_of_histograms)%values%binvar      = binvar
    histo_expanded(number_of_histograms)%values%binvar      = binvar
    histo_luminosity(number_of_histograms)%values%binvar    = binvar
    histo_resummed(number_of_histograms)%values%scale      = scale
    histo_expanded(number_of_histograms)%values%scale      = scale
    histo_luminosity(number_of_histograms)%values%scale    = scale

    histo_resummed(number_of_histograms)%values%right(:)   = bins
    histo_resummed(number_of_histograms)%values%weight(:)  = zero
    histo_resummed(number_of_histograms)%values%error(:)   = zero

    histo_expanded(number_of_histograms)%values%lo(:)     = zero
    histo_expanded(number_of_histograms)%values%nlo(:)    = zero
    histo_expanded(number_of_histograms)%values%nnlo(:)   = zero
    histo_expanded(number_of_histograms)%values%n3lo(:)   = zero

    histo_luminosity(number_of_histograms)%values%asympt  = zero
    histo_luminosity(number_of_histograms)%values%as      = zero
    histo_luminosity(number_of_histograms)%values%as2     = zero
    histo_luminosity(number_of_histograms)%values%as3     = zero
  end subroutine book_histogram


  ! routines to fill the histogram
  subroutine  fill_histogram(histo_index, obs, weight)
    real(dp), intent(in) :: weight, obs ! arrays of 7 entries (n. of weights computed in the shower)
    integer,  intent(in) :: histo_index
    !-----------------------------------------
    integer              :: i

    if (histo_resummed(histo_index)%values%scale /= scale_index) return
    if (histo_resummed(histo_index)%values%binvar /= binvar_index) return

    do i = 1, size(histo_resummed(histo_index)%values%right(:))
      if (obs <= histo_resummed(histo_index)%values%right(i)) then
        histo_resummed(histo_index)%values%weight(i) = histo_resummed(histo_index)%values%weight(i) + weight
        histo_resummed(histo_index)%values%error(i)  = histo_resummed(histo_index)%values%error(i)  + weight**2
      end if
    end do
    return
  end subroutine fill_histogram


  subroutine  fill_histogram_smooth(histo_index, obs, weight)
    real(dp), intent(in) :: weight, obs
    integer,  intent(in) :: histo_index
    !-----------------------------------------
    real(dp), parameter  :: theta_pow = 10._dp
    real(dp)             :: wgt_tmp
    integer              :: i, j

    if (histo_resummed(histo_index)%values%scale /= scale_index) return
    if (histo_resummed(histo_index)%values%binvar /= binvar_index) return

    do i = 1, size(histo_resummed(histo_index)%values%right(:))
      wgt_tmp = weight*(histo_resummed(histo_index)%values%right(i)/obs)**(two*theta_pow) &
      & /((histo_resummed(histo_index)%values%right(i)/obs)**(two*theta_pow) + one)
      histo_resummed(histo_index)%values%weight(i) = histo_resummed(histo_index)%values%weight(i) + wgt_tmp
      histo_resummed(histo_index)%values%error(i)  = histo_resummed(histo_index)%values%error(i)  + wgt_tmp**2
    end do
    return
  end subroutine fill_histogram_smooth


  subroutine  fill_histogram_array(histo_index, obs, weight)
    real(dp), intent(in) :: weight(:), obs(:) ! arrays of 7 entries (n. of weights computed in the shower)
    integer,  intent(in) :: histo_index
    !-----------------------------------------
    integer              :: i, j

    if (histo_resummed(histo_index)%values%scale /= scale_index) return
    if (histo_resummed(histo_index)%values%binvar /= binvar_index) return

    do i = 1, size(histo_resummed(histo_index)%values%right(:))
      do j = 1, size(obs)
        if (obs(j) == zero .or. weight(j) == zero) cycle
        if (obs(j) <= histo_resummed(histo_index)%values%right(i)) then
          histo_resummed(histo_index)%values%weight(i) = histo_resummed(histo_index)%values%weight(i) + weight(j)
          histo_resummed(histo_index)%values%error(i)  = histo_resummed(histo_index)%values%error(i)  + weight(j)**2
        end if
      end do
    end do
    return
  end subroutine fill_histogram_array


  subroutine  fill_histogram_array_smooth(histo_index, obs, weight, zeta_sp1, zeta_sp2)
    real(dp), intent(in)           :: weight(:), obs(:) ! arrays of 11 entries (n. of weights computed in the shower)
    integer,  intent(in)           :: histo_index
    real(dp), intent(in), optional :: zeta_sp1,zeta_sp2
    !-----------------------------------------
    real(dp)  :: tmp
    real(dp)  :: wgt_tmp(size(histo_resummed(histo_index)%values%right(:)))
    integer   :: i, j, theta_pow

    theta_pow = 10
    if (present(zeta_sp1).and.present(zeta_sp2)) then
       if ((zeta_sp1 < 1d-1) .or. (zeta_sp2 < 1d-1)) theta_pow = 5
    end if

    if (histo_resummed(histo_index)%values%scale /= scale_index) return
    if (histo_resummed(histo_index)%values%binvar /= binvar_index) return

    wgt_tmp = zero
    do j = 1, size(obs)
      if (obs(j) == zero .or. weight(j) == zero) cycle
      do i = 1, size(histo_resummed(histo_index)%values%right(:))
        if (obs(j) <= three*histo_resummed(histo_index)%values%right(i)) then
          if  ( obs(j) <= one/three*histo_resummed(histo_index)%values%right(i)) then
            wgt_tmp(i) = wgt_tmp(i) + weight(j)
          else
            tmp = (histo_resummed(histo_index)%values%right(i)/obs(j))**(2*theta_pow)
            wgt_tmp(i) = wgt_tmp(i) + weight(j)*tmp/(tmp + one)
          endif
        endif
      end do
    end do

    histo_resummed(histo_index)%values%weight = histo_resummed(histo_index)%values%weight + wgt_tmp
    histo_resummed(histo_index)%values%error  = histo_resummed(histo_index)%values%error  + wgt_tmp**2
    return
  end subroutine fill_histogram_array_smooth


  subroutine  fill_expansion(histo_index, weight)
    real(dp), intent(in) :: weight(0:,:)
    integer,  intent(in) :: histo_index
    !-----------------------------------------

    if (histo_resummed(histo_index)%values%scale /= scale_index) return
    if (histo_resummed(histo_index)%values%binvar /= binvar_index) return

    histo_expanded(histo_index)%values%lo(:)   = histo_expanded(histo_index)%values%lo(:)   + weight(0,:)
    histo_expanded(histo_index)%values%nlo(:)  = histo_expanded(histo_index)%values%nlo(:)  + weight(1,:)
    histo_expanded(histo_index)%values%nnlo(:) = histo_expanded(histo_index)%values%nnlo(:) + weight(2,:)
    histo_expanded(histo_index)%values%n3lo(:) = histo_expanded(histo_index)%values%n3lo(:) + weight(3,:)

    return
  end subroutine fill_expansion


  subroutine  fill_luminosity(histo_index, weight)
    real(dp), intent(in) :: weight(0:)
    integer,  intent(in) :: histo_index
    !-----------------------------------------

    if (histo_resummed(histo_index)%values%scale /= scale_index) return
    if (histo_resummed(histo_index)%values%binvar /= binvar_index) return

    histo_luminosity(histo_index)%values%asympt = histo_luminosity(histo_index)%values%asympt + weight(0)
    histo_luminosity(histo_index)%values%as     = histo_luminosity(histo_index)%values%as     + weight(1)
    histo_luminosity(histo_index)%values%as2    = histo_luminosity(histo_index)%values%as2    + weight(2)
    histo_luminosity(histo_index)%values%as3    = histo_luminosity(histo_index)%values%as3    + weight(3)

    return
  end subroutine fill_luminosity


  ! routine to normalise histograms
  subroutine normalise_histograms(number_of_born_events, number_of_shower_events) BIND(C)
    use iso_c_binding
    integer(c_int), intent(in) :: number_of_born_events, number_of_shower_events
    integer :: i
    real(dp), parameter :: convert_to_nb = 1d-6

    do i = 1, number_of_histograms
      histo_resummed(i)%values%weight(:) = histo_resummed(i)%values%weight(:)/number_of_born_events/number_of_shower_events*convert_to_nb
      histo_resummed(i)%values%error(:)  = sqrt(abs(histo_resummed(i)%values%error(:))) /number_of_born_events / number_of_shower_events*convert_to_nb
      ! histo_resummed(i)%values%error(:)  = histo_resummed(i)%values%error(:)/number_of_born_events/number_of_shower_events*convert_to_nb
      ! histo_resummed(i)%values%error(:)  = sqrt(abs(histo_resummed(i)%values%error(:) - histo_resummed(i)%values%weight(:)**2) &
      !   & /number_of_born_events/number_of_shower_events)*convert_to_nb

      histo_expanded(i)%values%lo(:)     = histo_expanded(i)%values%lo(:)/number_of_born_events*convert_to_nb
      histo_expanded(i)%values%nlo(:)    = histo_expanded(i)%values%nlo(:)/number_of_born_events*convert_to_nb
      histo_expanded(i)%values%nnlo(:)   = histo_expanded(i)%values%nnlo(:)/number_of_born_events*convert_to_nb
      histo_expanded(i)%values%n3lo(:)   = histo_expanded(i)%values%n3lo(:)/number_of_born_events*convert_to_nb

      histo_luminosity(i)%values%asympt  = histo_luminosity(i)%values%asympt/number_of_born_events*convert_to_nb
      histo_luminosity(i)%values%as      = histo_luminosity(i)%values%as/number_of_born_events*convert_to_nb
      histo_luminosity(i)%values%as2     = histo_luminosity(i)%values%as2/number_of_born_events*convert_to_nb
      histo_luminosity(i)%values%as3     = histo_luminosity(i)%values%as3/number_of_born_events*convert_to_nb
    end do

  end subroutine normalise_histograms


  ! routine to deallocate histograms
  subroutine deallocate_histograms() BIND(C)
    use iso_c_binding

    deallocate(histo_resummed)
    deallocate(histo_expanded)
    deallocate(histo_luminosity)

  end subroutine deallocate_histograms

end module histogram_routines
