! This module contains the routines that define the observables under study
module observables
  use momenta_and_recoil
  use types; use consts_dp
  use rad_tools

  implicit none
  private
  public :: compute_observable, dlgl, lndlbar, lndlbar2, lndlbar3, get_phi, azimuthal_zero, reset_caching

  !--- observables routines
  public :: obs_qt, obs_qtjet, obs_phistar 
  real(dp), parameter, public :: ktpow=1.0 ! this is the kt exponent in the observable's scaling
  ! The present implementation works only for ktpow=1. Other cases require major changes in the code

  !--- here add some caching to speed up the computation of observables
  logical, save, private :: caching
  type(scmom), save, private :: qt_tmp

contains

  subroutine compute_observable(name, observable, special1, special2, inclusive, res_order, icorr, eta_diff)
    real(dp), intent(out) :: observable
    logical,  intent(in)  :: special1, special2, inclusive
    integer,  intent(in)  :: res_order ! resummation order
    integer,     optional, intent(in) :: icorr
    real(dp),    optional, intent(in) :: eta_diff
    character(len=*), intent(in) :: name

    if (name == 'ptB') then
       observable=obs_qt(special1,special2) ! pt is inclusive so icorr is irrelevant
       
    elseif (name == 'ptj') then
       if (present(icorr).and.present(eta_diff)) then
          observable=obs_qtjet(special1,inclusive,res_order,icorr,eta_diff)
       else
          observable=obs_qtjet(special1,inclusive,res_order)
       end if
       
    elseif (name == 'JV_ptB') then
       ! compute pT of the leading jet
       if (present(icorr).and.present(eta_diff)) then
          observable=obs_qtjet(special1,inclusive,res_order,icorr,eta_diff)
       else
          observable=obs_qtjet(special1,inclusive,res_order)
       end if
       ! now check jet veto requirement and compute observable
       if (observable .le. cs%jet_veto) then
          observable=obs_qt(special1,special2)
       else
          ! set to a large value so it never gets binned
          observable=cs%rts 
       end if
       
    elseif (name == 'JV_ptj') then
       ! compute pT of the colour singlet
       observable=obs_qt(special1,special2)
       ! now impose the veto requirement
       if (observable .le. cs%jet_veto) then
          if (present(icorr).and.present(eta_diff)) then
             observable=obs_qtjet(special1,inclusive,res_order,icorr,eta_diff)
          else
             observable=obs_qtjet(special1,inclusive,res_order)
          end if
       else
          observable = cs%rts
       end if

    elseif (name == 'phi') then
       observable=obs_phistar(special1,special2)
       ! divide by the singlet's invariant mass
       if (cs%M.ne.zero) then
          observable=observable/cs%M
       else
          observable=zero
       end if
       
    elseif (name == 'pt2') then
       observable=obs_qt(special1,special2)/two ! this is ptB/2, useful for checks
       
    else
       write(*,*) name, ' compute observable error: observable not yet available'
       stop
    end if
  end subroutine compute_observable


  ! the following subroutine resets all cached quantitites
  subroutine reset_caching()
    caching = .true.
    call init_scmom(qt_tmp,zero)
  end subroutine reset_caching


  ! this computes the transverse momentum of the colour singlet
  function obs_qt(special1,special2) result(res)
    logical, intent(in) :: special1, special2
    real(dp) :: res
    integer :: i
    type(scmom) :: qttot

    call init_scmom(qttot,zero)
    ! inclusive sum over the transverse momenta
    if (caching) then
      do i=1,nems
       qttot = qttot + q(i)
      end do
      qt_tmp = qttot
      caching = .false.
    else
      qttot = qt_tmp
    end if

    if (special1) qttot = qttot + q(nems+1)
    if (special2) qttot = qttot + q(nems+2)
    res=qttot%ktmod
    return
  end function obs_qt

  ! this computes the phi* for DY lepton pair production
  function obs_phistar(special1,special2) result(res)
    logical, intent(in) :: special1, special2
    real(dp) :: res
    integer :: i

    res=zero
    do i=1,nems
       res = res + q(i)%kt(2)
    end do
    if (special1) res = res + q(nems+1)%kt(2)
    if (special2) res = res + q(nems+2)%kt(2)
    res=abs(res)
    return
  end function obs_phistar

  ! this computes the leading jet's transverse momentum
  ! valid for the generalized-kt family of jet algorithms
  ! dij = min(kti^(-2p),ktj^(-2p)) DRij^2/R^2
  ! diB = kti^(-2p)
  ! clustering between i and j if dij < min(diB,djB) --> DRij < R
  function obs_qtjet(special,inclusive,res_order,icorr,eta_diff) result(res)
    logical, intent(in) :: special, inclusive
    integer, intent(in) :: res_order
    integer, optional, intent(in) :: icorr
    real(dp),    optional, intent(in) :: eta_diff
    real(dp) :: res, dR, dphi
    integer :: i
    type(scmom), allocatable :: qjet(:)

    if (res_order==order_NLL) then
       ! no clustering between emissions
       if (.not.special) then
          i = sum(maxloc(q(1:nems)%ktmod))   ! exclude special emission
          res = q(i)%ktmod
       else
          i = sum(maxloc(q(1:nems+1)%ktmod)) ! include special emission
          res = q(i)%ktmod
       end if
    elseif (res_order==order_NNLL) then
       ! icorr and eta_diff needed
       if (.not.present(icorr).or..not.present(eta_diff)) then
          write(*,*) 'obs_qtjet error: pass icorr and eta_diff'
          stop
       end if
       
       allocate(qjet(nems+1))
       call init_scmom(qjet,zero)
       qjet(1:nems+1)=q(1:nems+1)
       if (inclusive) then
          ! work out the inclusive case:
          ! clustering of the two correlated emissions
          qjet(icorr) = qjet(icorr)+qjet(nems+1)
          i = sum(maxloc(qjet(1:nems)%ktmod)) ! exclude last emission
          res = qjet(i)%ktmod
       else
          ! work out the general NNLL case:
          dphi = acos(dot_product(qjet(nems+1)%kt,qjet(icorr)%kt)/(qjet(nems+1)%ktmod*qjet(icorr)%ktmod))
          dR = sqrt(eta_diff**2 + dphi**2)
          !--- check clustering condition
          if (dR < cs%jet_radius) then
             !--- cluster the correlated emissions
             qjet(icorr) = qjet(icorr)+qjet(nems+1)
             i = sum(maxloc(qjet(1:nems)%ktmod))
             res = qjet(i)%ktmod
          else
             i = sum(maxloc(qjet(1:nems+1)%ktmod))
             res = qjet(i)%ktmod
          end if
       end if
       deallocate(qjet)
    end if
    return
  end function obs_qtjet









  !---------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------
  !---------------------------------------------------------------------------------
  !--- the following routines are currently unusued, in general might be needed
  !--- for more exotic observables

  ! the following two functions define the azimuthal dependence and overall 
  ! normalisation of the observable under study
  ! v(k) = (kt)^a*exp(-b*eta)*dl*gl(phi)
  function dlgl(phi) result(res)
    real(dp), intent(in) :: phi !azimuthal angle
    real(dp) :: res
    if (cs%observable == 'ptB') then
       res = one
    elseif (cs%observable == 'ptj') then
       res = one
    elseif (cs%observable == 'phi') then
       res = abs(sin(phi))
    elseif (cs%observable == 'pt2') then
       res = half
    else
       write(*,*) cs%observable, ' observable not yet available'
       stop
    end if
  end function dlgl

  ! lndlbar = Integrate[Log[dl*gl(phi)]/(2*Pi),{phi,0,2*Pi}]
  function lndlbar() result(res)
    real(dp) :: res
    if (cs%observable == 'ptB') then
       res = zero
    elseif (cs%observable == 'ptj') then
       res = zero
    elseif (cs%observable == 'phi') then
       res = -log(two)
    elseif (cs%observable == 'pt2') then
       res = -log(two)
    else
       write(*,*) cs%observable, ' observable not yet available'
       stop
    end if
  end function lndlbar

  ! lndlbar2 = Integrate[Log^2[dl*gl(phi)]/(2*Pi),{phi,0,2*Pi}]
  function lndlbar2() result(res)
    real(dp) :: res
    if (cs%observable == 'ptB') then
       res = zero
    elseif (cs%observable == 'ptj') then
       res = zero
    elseif (cs%observable == 'phi') then
       res = pisq/12._dp+log(two)**2
    elseif (cs%observable == 'pt2') then
       res = log(two)**2
    else
       write(*,*) cs%observable, ' observable not yet available'
       stop
    end if
  end function lndlbar2

  ! lndlbar3 = Integrate[Log^3[dl*gl(phi)]/(2*Pi),{phi,0,2*Pi}]
  function lndlbar3() result(res)
    real(dp) :: res
    if (cs%observable == 'ptB') then
       res = zero
    elseif (cs%observable == 'ptj') then
       res = zero
    elseif (cs%observable == 'phi') then
       res = -pisq/4d0*log(two) -log(two)**3 - 3d0*zeta3/two
       stop "check lndlbar3() for phi..."
    elseif (cs%observable == 'pt2') then
       res = -log(two)**3
    else
       write(*,*) cs%observable, ' observable not yet available'
       stop
    end if
  end function lndlbar3

  ! the following function returns the azimuthal direction at which
  ! v(k) vanishes. Important for the computation of the Sudakov
  function azimuthal_zero() result(res)
    real(dp) :: res
    if (cs%observable == 'ptB') then
       res = 10._dp ! any number that is never sampled
    elseif (cs%observable == 'ptj') then
       res = 10._dp
    elseif (cs%observable == 'phi') then
       res = zero
    elseif (cs%observable == 'pt2') then
       res = 10._dp
    else
       write(*,*) cs%observable, ' observable not yet available'
       stop
    end if
  end function azimuthal_zero
  
  function get_phi(p) result(phi)
    type(scmom), intent(in) :: p
    real(dp) :: phi
    phi = acos(p%kt(1)/p%ktmod)
    if (p%kt(2)<zero) phi=-phi
    return
  end function get_phi

end module observables
