!========================================================
!--------------------------------------------------------
! Includes a module providing the parton luminosities
! and their convolutions with splitting and coefficient
! functions @ LO, NLO, NNLO, and a module providing
! the expansion of the luminosity factor up to order O(alphas^3)
! -------------------------------------------------------
!========================================================
module pdfs_tools
  use types; use consts_dp
  !! if using LHAPDF, rename a couple of hoppet functions which
  !! would otherwise conflict with LHAPDF
  use hoppet_v1, EvolvePDF_hoppet => EvolvePDF, InitPDF_hoppet => InitPDF
  use ew_parameters
  use rad_tools
  use interpolation_radish
  use warnings_and_errors

  implicit none

  type(dglap_holder), save, public :: dglap_h ! splitting function holder
  type(pdf_table),  save, public :: PDFs  ! parton densities
  type(pdf_table),  save, public :: C1_PDFs, C2_PDFs, G1_PDFs ! parton densities convoluted with coefficient functions
  type(grid_def), save, public :: grid
  type(split_mat), save, public :: C1_matrix, C2_matrix, G1_matrix, C1_plus_P, C2_plus_P
  type(split_mat), save, public ::  C1_x_P0_matrix, P0_matrix, P0_x_P0_matrix
  !!! PT+PM: the following is necessary to build the evolution operators
  type(pdf_table),  save, public :: P0_PDFs
  type(pdf_table),  save, public :: P0_NLO_PDFs
  type(pdf_table),  save, public :: P0_P0_PDFs
  type(pdf_table),  save, public :: C1_matrix_PDFs, C2_matrix_PDFs
  type(pdf_table),  save, public :: C1_P0_matrix_PDFs

  private

  public :: init_pdfs_from_LHAPDF
  public :: init_pdfs_custom_evolution
  public :: init_grid_and_dglap
  public :: get_pdfs, init_C_matxs, reset_dynamical_C_matxs
  public :: get_pdfs_Born, lumi_Born, lumi_NLL_Born, lumi_NNLL_Born, lumi_NNNLL_Born
  public :: dlumi_Born, P0_lumi_NLL, P0_P0_lumi_NLL
  public :: deallocate_grids, deallocate_C_matxs
  public :: inclusive_lumi_all_x, inclusive_lumi_at_x
  public :: inclusive_lumi_NLL_loc, inclusive_lumi_NNLL_loc, inclusive_lumi_NNNLL_loc
  public :: inclusive_P0_lumi_NLL, inclusive_P0_P0_lumi_NLL
  real(dp)            :: PDF_cutoff
  integer  :: nf_pdf
  integer, parameter  :: pdf_fit_order = -5
  real(dp), parameter :: dy = 0.10_dp, ymax = 15.0_dp, dlnlnQ = dy/4.0_dp

contains

  !=======================================================================================
  ! set up Hoppet's x grid, splitting functions and PDF tabulation
  ! object (just the memory allocation -- it gets filled below).
  subroutine init_grid_and_dglap(Qmax, fixed_flavour_scheme)
    use coefficient_functions_VH
    real(dp), intent(in), optional :: Qmax
    logical,  intent(in), optional :: fixed_flavour_scheme
    type(grid_def) ::  gdarray(4) ! grid

    ! build the PDF grid
    call InitGridDef(gdarray(4),dy/27.0_dp, 0.2_dp, order=pdf_fit_order)
    call InitGridDef(gdarray(3),dy/9.0_dp,  0.5_dp, order=pdf_fit_order)
    call InitGridDef(gdarray(2),dy/3.0_dp,  2.0_dp, order=pdf_fit_order)
    call InitGridDef(gdarray(1),dy,         ymax  , order=pdf_fit_order)
    call InitGridDef(grid,gdarray(1:4),locked=.true.)

    ! Using nf_pdf to set nf in HOPPET consistently.
    ! In the code, the HOPPET variable nf (real) is used in the analytical expressions
    ! and the HOPPET variable nf_int (integer) is used to set the dimension of matrices in flavour space

    call AllocSplitMat(grid, C1_matrix, nf_pdf)
    call AllocSplitMat(grid, C2_matrix, nf_pdf)
    call AllocSplitMat(grid, G1_matrix, nf_pdf)

    if (present(fixed_flavour_scheme)) then
       if (fixed_flavour_scheme) then
          !>> fixed flavour number scheme
          call qcd_SetNf(nf_pdf)
          call InitDglapHolder(grid, dglap_h,  factscheme = factscheme_MSbar, nloop = 3)
          call SetNfDglapHolder(dglap_h, nf_pdf)
       else
          !>> variable flavour number scheme [3 flavours]
          call dglap_Set_nnlo_splitting(nnlo_splitting_exact)
          call dglap_Set_nnlo_nfthreshold(nnlo_nfthreshold_exact)
          call InitDglapHolder(grid, dglap_h,  factscheme = factscheme_MSbar, nloop = 3, nflo = 3, nfhi = 6)
       end if
    else
       !>> fixed flavour number scheme
       call qcd_SetNf(nf_pdf)
       call InitDglapHolder(grid, dglap_h,  factscheme = factscheme_MSbar, nloop = 3)
       call SetNfDglapHolder(dglap_h, nf_pdf)
    end if

    call InitCoeffMatrix(grid, C1_matrix, C2_matrix, G1_matrix)

    if (present(Qmax)) then
       call AllocPdfTable(grid, PDFs, PDF_cutoff, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)
    else
       call AllocPdfTable(grid, PDFs, PDF_cutoff, 2d4, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)
    end if
  end subroutine init_grid_and_dglap

  !=======================================================================================
  !! initialise Hoppet's PDF table from LHAPDF
  subroutine init_pdfs_from_LHAPDF(pdf_name, pdf_set)
    character(len=*), intent(in) :: pdf_name
    integer,          intent(in) :: pdf_set
    !--------------------------------------------------
    real(dp) :: alphasQ, alphasPDF, ref_Q, PDF_cutoff2
    interface
       subroutine evolvePDF(x,Q,res)
         use types; implicit none
         real(dp), intent(in)  :: x,Q
         real(dp), intent(out) :: res(*)
       end subroutine evolvePDF
    end interface

    ! set up LHAPDF
    call InitPDFsetByName(trim(pdf_name))
    call InitPDF(pdf_set)
    call getNf(nf_pdf)

    ! read PDF cutoff
    call GetQ2min(pdf_set,PDF_cutoff2)
    PDF_cutoff = sqrt(PDF_cutoff2)

    call init_grid_and_dglap()

    ! sort out the coupling: get it from LHAPDF, which will have been initialised in InitPDF
    ref_Q = MZ
    alphasQ = alphasPDF(ref_Q)
    ! Initialise with variable number of active flavours
    call InitRunningCoupling(alfas=alphasQ, Q=ref_Q, nloop=3)

    call FillPdfTable_LHAPDF(PDFs, evolvePDF)
    !------------------------------------------------------------
  end subroutine init_pdfs_from_LHAPDF

  !=======================================================================================
  !! initialise Hoppet's PDF table by evolution from an LHAPDF input
  !! at some reference scale.
  subroutine init_pdfs_custom_evolution(pdf_name, pdf_set, rts, order)
    character(len=*),  intent(in) :: pdf_name
    integer,           intent(in) :: pdf_set
    real(dp),          intent(in) :: rts
    integer, optional, intent(in) :: order !>> this is the resummation order, DGLAP will be performed accordingly
    !--------------------------------------------------
    real(dp), pointer :: pdf_Q(:,:)
    real(dp) :: Qmax, ref_Q, alphas_Q, alphasPDF, PDF_cutoff2

    interface
       subroutine evolvePDF(x,Q,res)
         use types; implicit none
         real(dp), intent(in)  :: x,Q
         real(dp), intent(out) :: res(*)
       end subroutine evolvePDF
    end interface

    ! set up LHAPDF
    call InitPDFsetByName(trim(pdf_name))
    call InitPDF(pdf_set)
    call getNf(nf_pdf)

    if (present(order)) then
       write(0,*) "WARNING: performing a DGLAP evolution with", nf_int, " active flavours and order", order
    else
       write(0,*) "WARNING: performing a DGLAP evolution with", nf_int, " active flavours and order", 3
    end if

    ! read PDF cutoff
    call GetQ2min(pdf_set,PDF_cutoff2)
    PDF_cutoff = sqrt(PDF_cutoff2)

    Qmax = max(2e4_dp, rts)
    call init_grid_and_dglap(Qmax)
    call AllocPDF(grid, pdf_Q)

    ref_Q = MZ  !>> do backward evolution with 5 flavours from MZ to the low scale
    alphas_Q = alphasPDF(ref_Q)
    call InitPDF_LHAPDF(grid, pdf_Q, evolvePDF, ref_Q)

    ! set up a running coupling, set up the PDF tabulation and perform the evolution
    if (present(order)) then
       call InitRunningCoupling(alfas=alphas_Q, Q=ref_Q, nloop=order, fixnf=nf_int)
       call EvolvePdfTable(PDFs, ref_Q, pdf_Q, dglap_h, ash_global, nloop=order)
    else
       call InitRunningCoupling(alfas=alphas_Q, Q=ref_Q, nloop=3, fixnf=nf_int)
       call EvolvePdfTable(PDFs, ref_Q, pdf_Q, dglap_h, ash_global)
    end if
    !------------------------------------------------------------
  end subroutine init_pdfs_custom_evolution

  !=======================================================================================
  ! initialise coefficient functions used for NNLL luminosity
  subroutine init_C_matxs()
    implicit none
    real(dp) :: Qmax, muF_ref
    integer  :: iQ

    !----------------------------------------------
    call AllocSplitMat(grid, P0_matrix, nf_int)
    call InitSplitMat(P0_matrix, dglap_h%P_LO)

    call AllocSplitMat(grid, P0_x_P0_matrix, nf_int)
    call AllocSplitMat(grid, C1_x_P0_matrix, nf_int)
    call SetToConvolution(P0_x_P0_matrix, dglap_h%P_LO, dglap_h%P_LO)
    call SetToConvolution(C1_x_P0_matrix, C1_matrix, dglap_h%P_LO)

    ! Initialise coefficient functions used in the inclusive case

    ! O(as) coefficient functions
    call InitSplitMat(C1_plus_P, C1_matrix)
    ! add resummation scale dependence
    call AddWithCoeff(C1_plus_P, dglap_h%P_LO, -cs%ln_muF2_M2+cs%ln_Q2_M2)

    ! O(as^2) coefficient functions
    call InitSplitMat(C2_plus_P, C2_matrix)
    ! add resummation scale dependence
    call AddWithCoeff(C2_plus_P, dglap_h%P_LO, pi*beta0*(-cs%ln_muF2_M2+cs%ln_Q2_M2)**2 &
          & - two*pi*beta0*(-cs%ln_muF2_M2+cs%ln_Q2_M2)*cs%ln_Q2_muR2)
    call AddWithCoeff(C2_plus_P, dglap_h%P_NLO, -cs%ln_muF2_M2+cs%ln_Q2_M2)
    call AddWithCoeff(C2_plus_P, P0_x_P0_matrix, half*(-cs%ln_muF2_M2+cs%ln_Q2_M2)**2)
    call AddWithCoeff(C2_plus_P, C1_x_P0_matrix, -cs%ln_muF2_M2+cs%ln_Q2_M2)
    call AddWithCoeff(C2_plus_P, C1_matrix, -two*pi*beta0*cs%ln_Q2_muR2)

    ! In the following create a pseudo-pdf which contains the convolution between the
    ! coefficient functions and the pdfs
    ! build the PDF grid: WARNING: the following parameters are to be set as in init_grid_and_dglap
    Qmax = max(2e4_dp, cs%rts)
    muF_ref=PDF_cutoff

    call AllocPdfTable(grid, C1_PDFs, muF_ref, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)
    call AllocPdfTable(grid, C2_PDFs, muF_ref, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)
    call AllocPdfTable(grid, C1_matrix_PDFs, muF_ref, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)
    call AllocPdfTable(grid, C2_matrix_PDFs, muF_ref, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)
    call AllocPdfTable(grid, G1_PDFs, muF_ref, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)

    call AllocPdfTable(grid, C1_P0_matrix_PDFs, muF_ref, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)
    call AllocPdfTable(grid, P0_PDFs, muF_ref, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)
    call AllocPdfTable(grid, P0_NLO_PDFs, muF_ref, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)
    call AllocPdfTable(grid, P0_P0_PDFs, muF_ref, Qmax, dlnlnQ=dlnlnQ, freeze_at_Qmin = .true.)

    ! build useful convolution between pdfs and coefficient functions
    do iQ=0, PDFs%nQ
      ! Start from O(as)
      C1_matrix_PDFs%tab(:,:,iQ) = C1_matrix    .conv. PDFs%tab(:,:,iQ)
      P0_PDFs%tab(:,:,iQ)        = dglap_h%P_LO .conv. PDFs%tab(:,:,iQ)
      P0_NLO_PDFs%tab(:,:,iQ)        = dglap_h%P_NLO .conv. PDFs%tab(:,:,iQ)

      C1_PDFs%tab(:,:,iQ)        = C1_matrix_PDFs%tab(:,:,iQ) + P0_PDFs%tab(:,:,iQ) * (-cs%ln_muF2_M2+cs%ln_Q2_M2) ! C1_plus_P .conv. PDFs%tab(:,:,iQ)

      ! Now handle O(as^2)
      C2_matrix_PDFs%tab(:,:,iQ)        = C2_matrix      .conv. PDFs%tab(:,:,iQ)
      P0_P0_PDFs%tab(:,:,iQ)            = dglap_h%P_LO .conv. (dglap_h%P_LO .conv. PDFs%tab(:,:,iQ))
      !P0_x_P0_matrix .conv. PDFs%tab(:,:,iQ)
      C1_P0_matrix_PDFs%tab(:,:,iQ)     = C1_x_P0_matrix .conv. PDFs%tab(:,:,iQ)

      C2_PDFs%tab(:,:,iQ)        = C2_matrix_PDFs%tab(:,:,iQ) &
            & + P0_PDFs%tab(:,:,iQ) * (pi*beta0*(-cs%ln_muF2_M2+cs%ln_Q2_M2)**2 &
            & - two*pi*beta0*(-cs%ln_muF2_M2+cs%ln_Q2_M2)*cs%ln_Q2_muR2) &
            & + P0_NLO_PDFs%tab(:,:,iQ) * (-cs%ln_muF2_M2+cs%ln_Q2_M2) &
            & + P0_P0_PDFs%tab(:,:,iQ) * (half*(-cs%ln_muF2_M2+cs%ln_Q2_M2)**2) &
            & + C1_P0_matrix_PDFs%tab(:,:,iQ) * (-cs%ln_muF2_M2+cs%ln_Q2_M2) &
            & + C1_matrix_PDFs%tab(:,:,iQ) * (-two*pi*beta0*cs%ln_Q2_muR2)

      G1_PDFs%tab(:,:,iQ)        = G1_matrix .conv. PDFs%tab(:,:,iQ)

    end do
    return

  end subroutine init_C_matxs

  !=======================================================================================
  ! initialise coefficient functions used for NNLL luminosity
  subroutine reset_dynamical_C_matxs()
    implicit none
    integer  :: iQ
    !----------------------------------------------
    ! update the various grids
    do iQ=0, PDFs%nQ
      C1_PDFs%tab(:,:,iQ)        = C1_matrix_PDFs%tab(:,:,iQ) + P0_PDFs%tab(:,:,iQ) * (-cs%ln_muF2_M2+cs%ln_Q2_M2) ! C1_plus_P .conv. PDFs%tab(:,:,iQ)

      C2_PDFs%tab(:,:,iQ)       = C2_matrix_PDFs%tab(:,:,iQ) &
            & + P0_PDFs%tab(:,:,iQ) * (pi*beta0*(-cs%ln_muF2_M2+cs%ln_Q2_M2)**2 &
            & - two*pi*beta0*(-cs%ln_muF2_M2+cs%ln_Q2_M2)*cs%ln_Q2_muR2) &
            & + P0_NLO_PDFs%tab(:,:,iQ) * (-cs%ln_muF2_M2+cs%ln_Q2_M2) &
            & + P0_P0_PDFs%tab(:,:,iQ) * (half*(-cs%ln_muF2_M2+cs%ln_Q2_M2)**2) &
            & + C1_P0_matrix_PDFs%tab(:,:,iQ) * (-cs%ln_muF2_M2+cs%ln_Q2_M2) &
            & + C1_matrix_PDFs%tab(:,:,iQ) * (-two*pi*beta0*cs%ln_Q2_muR2)

    end do
    return
  end subroutine reset_dynamical_C_matxs


  !======================================================================
  ! return the two incoming PDFs evaluated at scale muF, taking into
  ! account whether the collider is pp or ppbar
  subroutine get_pdfs(muF, collider, pdf1, pdf2)
    real(dp),         intent(in)  :: muF
    character(len=*), intent(in)  :: collider
    real(dp),         intent(out) :: pdf1(:,-6:), pdf2(:,-6:)
    !---------------------------------------------------------

    call EvalPdfTable_Q(PDFs, muF, pdf1)
    select case(trim(collider))
    case("pp")
       pdf2 = pdf1
    case("ppbar")
       pdf2(:,-6:6) = pdf1(:,6:-6:-1)
       pdf2(:,7)    = pdf1(:,7)          ! index 7 contains info on representation in flavour space
    case default
       call wae_error("unrecognized collider: "//collider)
    end select

    !--- now select the desired flavours (in human representation)
    if (flavour_breakdown) then
       ! First PDF
       if (cs%flavour1 == 'q') pdf1(:,-6:0)=zero
       if (cs%flavour1 == 'x') pdf1(:,0:6)=zero
       if (cs%flavour1 == 'g') then
          pdf1(:,-6:-1)=zero
          pdf1(:,1:6)=zero
       end if

       ! Second PDF
       if (cs%flavour2 == 'q') pdf2(:,-6:0)=zero
       if (cs%flavour2 == 'x') pdf2(:,0:6)=zero
       if (cs%flavour2 == 'g') then
          pdf2(:,-6:-1)=zero
          pdf2(:,1:6)=zero
       end if
    end if
  end subroutine get_pdfs




!========================================================
!--------------------------------------------------------
! Includes routines providing the parton luminosities
! and their convolutions with splitting and coefficient
! functions @ LO, NLO, NNLO for a given Born event
! -------------------------------------------------------
!========================================================


  !======================================================================
  ! return the two incoming PDFs evaluated at scale muF, and x, taking into
  ! account whether the collider is pp or ppbar
  subroutine get_pdfs_Born(muF, x1, x2, collider, pdf1, pdf2)
    real(dp),         intent(in)  :: muF, x1, x2
    character(len=*), intent(in)  :: collider
    real(dp),         intent(out) :: pdf1(-6:), pdf2(-6:)
    !---------------------------------------------------------

    pdf1=zero
    pdf2=zero
    ! evaluate pdfs at a given scale and x
    call EvalPdfTable_xQ(PDFs, x1, muF, pdf1)
    call EvalPdfTable_xQ(PDFs, x2, muF, pdf2)

    select case(trim(collider))
    case("pp")
       pdf2 = pdf2
    case("ppbar")
       pdf2(-6:6) = pdf2(6:-6:-1)
       pdf2(7)    = pdf2(7)          ! index 7 contains info on representation in flavour space
    case default
       call wae_error("unrecognized collider: "//collider)
    end select
    ! Hoppet returns x*f(x)
    pdf1 = pdf1/x1
    pdf2 = pdf2/x2

  end subroutine get_pdfs_Born



  !======================================================================
  ! this works out the partonic luminosity for the LO of process proc;
  !gives a luminosity based on the two input PDFs, which are evaluated at
  ! a given muF and x depending on the Born kinematics
  ! Remember this reads f(x) as an input instead of the conventional x*f(x) of hoppet
  function lumi_Born(pdf1, pdf2, msqB) result(res)
    real(dp),                     intent(in) :: pdf1(-6:), pdf2(-6:)
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int) ! Matrix containing the Born matrix elements squared
    real(dp)                     :: res
    !----------------------------
    integer :: i, j

    ! build parton luminosity
    res=zero
    do i= -nf_int, nf_int
       do j= -nf_int, nf_int
          res = res + msqB(i,j)*pdf1(i)*pdf2(j)
       end do
    end do

  end function lumi_Born

  !======================================================================
  ! returns the luminosity that enters the cross section at NLL, which
  ! is to be evaluated separately for each of the input muF values
  function lumi_NLL_Born(muF, lambda, x1, x2, msqB, msqV) result(res)
    real(dp),                  intent(in) :: muF, x1, x2
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int) ! Matrix containing the Born matrix elements squared
    real(dp), intent(in) :: msqV(-nf_int:nf_int,-nf_int:nf_int)
    real(dp) :: C1_conv_pdf1(-6:7), C1_conv_pdf2(-6:7)
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(-6:7), pdf2(-6:7)
    integer  :: i
    real(dp), intent(in) :: lambda
    real(dp) :: as2pi_mu

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    if(cs%prime) then
      ! get the pdf grids to compute convolutions with the coefficient functions
      C1_conv_pdf1=zero
      C1_conv_pdf2=zero
      call EvalPdfTable_xQ(C1_PDFs, x1, muF, C1_conv_pdf1)
      call EvalPdfTable_xQ(C1_PDFs, x2, muF, C1_conv_pdf2)

      ! Hoppet returns x*f(x)
      C1_conv_pdf1 = C1_conv_pdf1/x1
      C1_conv_pdf2 = C1_conv_pdf2/x2

      ! get the pdfs evaluated at a specific x according to the Born kinematics
      call get_pdfs_Born(muF, x1, x2, cs%collider, pdf1, pdf2)
      ! get the running coupling at scale muR = muR_default * exp(-L)
      as2pi_mu = cs%as2pi / (1-2*lambda)

      ! correct the luminosities and get no "spurious" as2pi_mu^2 term
      res = lumi_Born(pdf1, pdf2, msqB + cs%as2pi*msqV)
      res = res + as2pi_mu*lumi_Born(C1_conv_pdf1, pdf2, msqB)
      res = res + as2pi_mu*lumi_Born(pdf1, C1_conv_pdf2, msqB)
    else
      call get_pdfs_Born(muF, x1, x2, cs%collider, pdf1, pdf2)
      res = lumi_Born(pdf1, pdf2, msqB)
    endif

  end function lumi_NLL_Born

  !=======================================================================================
  ! The following two routines implement the additional pieces in our Eq 3.15 involving the convolution
  ! of the NLL luminosity with one and two splitting functions
  function P0_lumi_NLL(muF, lambda, x1, x2, msqB) result(res)
    real(dp), intent(in) :: muF, x1, x2
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int) ! Matrix containing the Born matrix elements squared
    real(dp), intent(in) :: lambda
    real(dp)             :: res
    !----------------------------------------------
    real(dp) :: pdf1(-6:7), pdf2(-6:7)
    real(dp) :: P0_conv_pdf1(-6:7), P0_conv_pdf2(-6:7)
    integer  :: i
    real(dp) :: as2pi

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    ! get the pdf grids to compute convolutions with the coefficient functions
    P0_conv_pdf1=zero
    P0_conv_pdf2=zero
    call EvalPdfTable_xQ(P0_PDFs, x1, muF, P0_conv_pdf1)
    call EvalPdfTable_xQ(P0_PDFs, x2, muF, P0_conv_pdf2)
    ! Hoppet returns x*f(x)
    P0_conv_pdf1 = P0_conv_pdf1/x1
    P0_conv_pdf2 = P0_conv_pdf2/x2


    call get_pdfs_Born(muF, x1, x2, cs%collider, pdf1, pdf2)
    res = lumi_Born(pdf1, P0_conv_pdf2, msqB) + lumi_Born(P0_conv_pdf1, pdf2, msqB)
    ! insert coupling and beta function factor
    res = res * two*beta0/pi*cs%alphas_muR**2/(one-two*lambda)**2

  end function P0_lumi_NLL

  !=======================================================================================
  function P0_P0_lumi_NLL(muF, lambda, x1, x2, msqB) result(res)
    real(dp), intent(in) :: muF, x1, x2
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int) ! Matrix containing the Born matrix elements squared
    real(dp), intent(in) :: lambda
    real(dp)             :: res
    !----------------------------------------------
    real(dp) :: pdf1(-6:7), pdf2(-6:7)
    real(dp) :: P0_P0_conv_pdf1(-6:7), P0_P0_conv_pdf2(-6:7)
    real(dp) :: P0_conv_pdf1(-6:7), P0_conv_pdf2(-6:7)
    integer  :: i
    real(dp) :: as2pi

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    ! get the pdf grids to compute convolutions with the coefficient functions
    P0_P0_conv_pdf1=zero
    P0_P0_conv_pdf2=zero
    call EvalPdfTable_xQ(P0_P0_PDFs, x1, muF, P0_P0_conv_pdf1)
    call EvalPdfTable_xQ(P0_P0_PDFs, x2, muF, P0_P0_conv_pdf2)
    ! Hoppet returns x*f(x)
    P0_P0_conv_pdf1 = P0_P0_conv_pdf1/x1
    P0_P0_conv_pdf2 = P0_P0_conv_pdf2/x2

    P0_conv_pdf1=zero
    P0_conv_pdf2=zero
    call EvalPdfTable_xQ(P0_PDFs, x1, muF, P0_conv_pdf1)
    call EvalPdfTable_xQ(P0_PDFs, x2, muF, P0_conv_pdf2)
    ! Hoppet returns x*f(x)
    P0_conv_pdf1 = P0_conv_pdf1/x1
    P0_conv_pdf2 = P0_conv_pdf2/x2


    call get_pdfs_Born(muF, x1, x2, cs%collider, pdf1, pdf2)
    res = lumi_Born(pdf1, P0_P0_conv_pdf2, msqB) + lumi_Born(P0_P0_conv_pdf1, pdf2, msqB)
    res = res + two*lumi_Born(P0_conv_pdf1, P0_conv_pdf2, msqB)
    ! insert coupling and beta function factor
    res = res * cs%alphas_muR**2/pi**2/(one-two*lambda)**2

  end function P0_P0_lumi_NLL

  !======================================================================
  ! returns the luminosity that enters the cross section at NNLL,
  ! including C1 terms. It is evaluated at each of the supplied muF
  ! values, and the user must supply to corresponding lambda values,
  ! which are of relevance for establishing the coupling that
  ! multiplies the C1 pieces.
  function lumi_NNLL_Born(muF, lambda, x1, x2, msqB, msqV1, msqV2) result(res)
    real(dp),                  intent(in) :: muF, lambda, x1, x2
    ! Matrix containing the Born and virtual matrix elements in flavour space
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int), msqV1(-nf_int:nf_int,-nf_int:nf_int), msqV2(-nf_int:nf_int,-nf_int:nf_int)
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(-6:7), pdf2(-6:7)
    real(dp) :: C1_conv_pdf1(-6:7), C1_conv_pdf2(-6:7)
    real(dp) :: C2_conv_pdf1(-6:7), C2_conv_pdf2(-6:7)
    real(dp) :: G1_conv_pdf1(-6:7), G1_conv_pdf2(-6:7)
    real(dp) :: as2pi_mu
    integer  :: i
    type(pdf_rep) :: pdfrep
    real(dp) :: pdfgrid1(0:grid%ny,-6:7), pdfgrid2(0:grid%ny,-6:7)

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    ! get the pdf grids to compute convolutions with the coefficient functions
    C1_conv_pdf1=zero
    C1_conv_pdf2=zero
    call EvalPdfTable_xQ(C1_PDFs, x1, muF, C1_conv_pdf1)
    call EvalPdfTable_xQ(C1_PDFs, x2, muF, C1_conv_pdf2)

    ! Hoppet returns x*f(x)
    C1_conv_pdf1 = C1_conv_pdf1/x1
    C1_conv_pdf2 = C1_conv_pdf2/x2

    ! get the pdfs evaluated at a specific x according to the Born kinematics
    call get_pdfs_Born(muF, x1, x2, cs%collider, pdf1, pdf2)
    ! get the running coupling at scale muR = muR_default * exp(-L)
    if (cs%prime) then
       as2pi_mu = cs%as2pi / (1-two*lambda) * (one &
            & - (twopi*cs%as2pi) / (1-two*lambda) * beta1/beta0 * log(one-two*lambda))
    else
       as2pi_mu = cs%as2pi / (1-2*lambda)
    end if

    ! correct the luminosities and get no "spurious" as2pi_mu^2 term
    res = lumi_Born(pdf1, pdf2, msqB + cs%as2pi*msqV1)
    res = res + as2pi_mu*lumi_Born(C1_conv_pdf1, pdf2, msqB)
    res = res + as2pi_mu*lumi_Born(pdf1, C1_conv_pdf2, msqB)

    if (cs%prime) then

      C2_conv_pdf1=zero
      C2_conv_pdf2=zero
      call EvalPdfTable_xQ(C2_PDFs, x1, muF, C2_conv_pdf1)
      call EvalPdfTable_xQ(C2_PDFs, x2, muF, C2_conv_pdf2)

      G1_conv_pdf1=zero
      G1_conv_pdf2=zero
      call EvalPdfTable_xQ(G1_PDFs, x1, muF, G1_conv_pdf1)
      call EvalPdfTable_xQ(G1_PDFs, x2, muF, G1_conv_pdf2)

      ! Hoppet returns x*f(x)
      C2_conv_pdf1 = C2_conv_pdf1/x1
      C2_conv_pdf2 = C2_conv_pdf2/x2

      G1_conv_pdf1 = G1_conv_pdf1/x1
      G1_conv_pdf2 = G1_conv_pdf2/x2

      ! Add the O(as^2) pieces
      ! get the one-loop running coupling at scale muR = muR_default * exp(-L)
      as2pi_mu = cs%as2pi / (1-2*lambda)
      res = res + as2pi_mu**2*lumi_Born(C2_conv_pdf1, pdf2, msqB)
      res = res + as2pi_mu**2*lumi_Born(pdf1, C2_conv_pdf2, msqB)

      ! add C1 x C1
      res = res + as2pi_mu**2*lumi_Born(C1_conv_pdf1, C1_conv_pdf2, msqB)
      ! add G1 x G1
      res = res + as2pi_mu**2*lumi_Born(G1_conv_pdf1, G1_conv_pdf2, msqB)
      ! add H2 and H1 * C1
      res = res + cs%as2pi**2*lumi_Born(pdf1, pdf2, msqV2)
      res = res + cs%as2pi*as2pi_mu*(lumi_Born(C1_conv_pdf1, pdf2, msqV1) &
        & + lumi_Born(pdf1, C1_conv_pdf2, msqV1))
    endif

  end function lumi_NNLL_Born

  !=======================================================================================
  ! The following routine returns the N3LL luminosity
  function lumi_NNNLL_Born(muF, lambda, x1, x2, msqB, msqV1, msqV2) result(res)
    real(dp),                  intent(in) :: muF, lambda, x1, x2
    ! Matrix containing the Born and virtual matrix elements in flavour space
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int), msqV1(-nf_int:nf_int,-nf_int:nf_int), msqV2(-nf_int:nf_int,-nf_int:nf_int)
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(-6:7), pdf2(-6:7)
    real(dp) :: C1_conv_pdf1(-6:7), C1_conv_pdf2(-6:7)
    real(dp) :: C2_conv_pdf1(-6:7), C2_conv_pdf2(-6:7)
    real(dp) :: G1_conv_pdf1(-6:7), G1_conv_pdf2(-6:7)
    real(dp) :: as2pi_mu
    integer  :: i
    type(pdf_rep) :: pdfrep
    real(dp) :: pdfgrid1(0:grid%ny,-6:7), pdfgrid2(0:grid%ny,-6:7)

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    ! get the pdf grids to compute convolutions with the coefficient functions
    C1_conv_pdf1=zero
    C1_conv_pdf2=zero
    call EvalPdfTable_xQ(C1_PDFs, x1, muF, C1_conv_pdf1)
    call EvalPdfTable_xQ(C1_PDFs, x2, muF, C1_conv_pdf2)

    C2_conv_pdf1=zero
    C2_conv_pdf2=zero
    call EvalPdfTable_xQ(C2_PDFs, x1, muF, C2_conv_pdf1)
    call EvalPdfTable_xQ(C2_PDFs, x2, muF, C2_conv_pdf2)

    G1_conv_pdf1=zero
    G1_conv_pdf2=zero
    call EvalPdfTable_xQ(G1_PDFs, x1, muF, G1_conv_pdf1)
    call EvalPdfTable_xQ(G1_PDFs, x2, muF, G1_conv_pdf2)


    ! Hoppet returns x*f(x)
    C1_conv_pdf1 = C1_conv_pdf1/x1
    C1_conv_pdf2 = C1_conv_pdf2/x2

    C2_conv_pdf1 = C2_conv_pdf1/x1
    C2_conv_pdf2 = C2_conv_pdf2/x2

    G1_conv_pdf1 = G1_conv_pdf1/x1
    G1_conv_pdf2 = G1_conv_pdf2/x2

    ! get the pdfs evaluated at a specific x according to the Born kinematics
    call get_pdfs_Born(muF, x1, x2, cs%collider, pdf1, pdf2)
    ! get the two-loop running coupling at scale muR = muR_default * exp(-L)
    as2pi_mu = cs%as2pi / (1-two*lambda) * (one &
         & - (twopi*cs%as2pi) / (1-two*lambda) * beta1/beta0 * log(one-two*lambda))

    ! correct the luminosities and get no "spurious" as2pi_mu^2 term
    res = lumi_Born(pdf1, pdf2, msqB + cs%as2pi*msqV1)
    res = res + as2pi_mu*lumi_Born(C1_conv_pdf1, pdf2, msqB)
    res = res + as2pi_mu*lumi_Born(pdf1, C1_conv_pdf2, msqB)

    ! Add the O(as^2) pieces
    ! get the one-loop running coupling at scale muR = muR_default * exp(-L)
    as2pi_mu = cs%as2pi / (1-two*lambda)

    res = res + as2pi_mu**2*lumi_Born(C2_conv_pdf1, pdf2, msqB)
    res = res + as2pi_mu**2*lumi_Born(pdf1, C2_conv_pdf2, msqB)

    ! add C1 x C1
    res = res + as2pi_mu**2*lumi_Born(C1_conv_pdf1, C1_conv_pdf2, msqB)
    ! add G1 x G1
    res = res + as2pi_mu**2*lumi_Born(G1_conv_pdf1, G1_conv_pdf2, msqB)
    ! add H2 and H1 * C1
    res = res + cs%as2pi**2*lumi_Born(pdf1, pdf2, msqV2)
    res = res + cs%as2pi*as2pi_mu*(lumi_Born(C1_conv_pdf1, pdf2, msqV1) &
         & + lumi_Born(pdf1, C1_conv_pdf2, msqV1))

  end function lumi_NNNLL_Born

  !=======================================================================================
  ! returns the numerical first and second derivative of the interpolated luminosity
  subroutine dlumi_Born(tt,lumi,dlumi,d2lumi)
    implicit none
    real(dp), intent(in)            :: tt(:), lumi(:)
    real(dp), intent(out)           :: dlumi(size(tt))
    real(dp), intent(out), optional :: d2lumi(size(tt))
    real(dp) :: dlumi_up, dlumi_down, histep, dy_error
    integer  :: i

    dlumi = zero
    dlumi_up = zero; dlumi_down = zero

    do i=1,size(tt)
       histep=1d-3 ! initial step for evaluation of numerical derivative
       do
          if (tt(i) - histep > zero) exit
          histep = histep/two
       end do
       ! evaluate the luminosity at the points that enter the discrete derivative
       call interpolate(tt(:),lumi(:),tt(i)-histep,dlumi_down,dy_error,3)
       call interpolate(tt(:),lumi(:),tt(i)+histep,dlumi_up,dy_error,3)
       dlumi(i) = (dlumi_up - dlumi_down)/(two*histep)
    end do

    if (present(d2lumi)) then
      d2lumi = zero
      do i=1,size(tt)
        histep=1d-3 ! initial step for evaluation of numerical derivative
        do
          if (tt(i) - histep > zero) exit
          histep = histep/two
        end do
        ! second derivative
        call interpolate(tt(:),dlumi(:),tt(i)-histep,dlumi_down,dy_error,3)
        call interpolate(tt(:),dlumi(:),tt(i)+histep,dlumi_up,dy_error,3)
        d2lumi(i) = (dlumi_up - dlumi_down)/(two*histep)
      end do
    end if
    return
  end subroutine dlumi_Born

  !======================================================================
  ! Functions needed for the inclusive case
  !======================================================================

  !=======================================================================================
  function inclusive_lumi_all_x(pdf1, pdf2) result(res)
    real(dp),                     intent(in) :: pdf1(:,-6:), pdf2(:,-6:)
    real(dp)                     :: res(size(pdf1,1))
    !----------------------------
    integer i

    select case(trim(cs%proc))
    case("H")
       res = PartonLuminosity(grid, pdf1(:,iflv_g), pdf2(:,iflv_g))
    case('DY')
       res = zero
       do i = 1, nf_int
          res = res + gv2_ga2(i) * PartonLuminosity(grid, pdf1(:, i), pdf2(:,-i))
          res = res + gv2_ga2(i) * PartonLuminosity(grid, pdf1(:,-i), pdf2(:, i))
       end do
    case default
      call wae_error("unrecognized process: "//trim(cs%proc))
    end select
    res = res * cs%norm_fb ! Normalise in order to get the cross section in [fb]
  end function inclusive_lumi_all_x

  function inclusive_lumi_at_x(pdf1, pdf2) result(res)
    real(dp),                     intent(in) :: pdf1(:,-6:), pdf2(:,-6:)
    real(dp)                     :: res
    !--------------------------------------
    res = inclusive_lumi_all_x(pdf1, pdf2) .atx. (cs%M2_rts2 .with. grid)
  end function inclusive_lumi_at_x

  !======================================================================
  ! returns the luminosity that enters the cross section at NLL, which
  ! is evaluated for a given scale muF
  function inclusive_lumi_NLL_loc(muF, lambda) result(res)
    real(dp),                  intent(in) :: muF, lambda
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    real(dp) :: C1_x_pdf1(0:grid%ny,-6:7), C1_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: as2pi_mu, HH1

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    if(cs%prime) then
      if (cs%loop_mass .ne. 'none') then
        HH1 = Hm(1)
      else
        HH1 = H(1)
      end if

      call get_pdfs(muF, cs%collider, pdf1, pdf2)

      ! get the running coupling at scale muR = muR_default * exp(-L)
      as2pi_mu = cs%as2pi / (1-2*lambda)
      ! get the luminosities w/o spurious O(as^3) terms
      ! O(as^0) contribution
      res = inclusive_lumi_at_x(pdf1, pdf2)

      ! O(as) contribution
      C1_x_pdf1 = (C1_plus_P .conv. pdf1)
      C1_x_pdf2 = (C1_plus_P .conv. pdf2)
      res = res + as2pi_mu*inclusive_lumi_at_x(C1_x_pdf1, pdf2)
      res = res + as2pi_mu*inclusive_lumi_at_x(pdf1, C1_x_pdf2)

      ! add virtual corrections
      res = res + cs%as2pi*HH1*inclusive_lumi_at_x(pdf1, pdf2)
    else
      call get_pdfs(muF, cs%collider, pdf1, pdf2)
      res = inclusive_lumi_at_x(pdf1, pdf2)
    endif

  end function inclusive_lumi_NLL_loc

  !=======================================================================================
  ! same as above, but for the NNLL luminosity
  ! virtual corrections are part of the luminosity
  function inclusive_lumi_NNLL_loc(muF, lambda) result(res)
    real(dp),                  intent(in) :: muF, lambda
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    real(dp) :: C1_x_pdf1(0:grid%ny,-6:7), C1_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: C2_x_pdf1(0:grid%ny,-6:7), C2_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: G1_x_pdf1(0:grid%ny,-6:7), G1_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: as2pi_mu, HH1

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    if (cs%loop_mass .ne. 'none') then
       HH1 = Hm(1)
    else
       HH1 = H(1)
    end if

    call get_pdfs(muF, cs%collider, pdf1, pdf2)

    ! get the running coupling at scale muR = muR_default * exp(-L)
    if (cs%prime) then
       as2pi_mu = cs%as2pi / (1-two*lambda) * (one &
            & - (twopi*cs%as2pi) / (1-two*lambda) * beta1/beta0 * log(one-two*lambda))
    else
       as2pi_mu = cs%as2pi / (1-2*lambda)
    end if

    ! get the luminosities w/o spurious O(as^3) terms
    ! O(as^0) contribution
    res = inclusive_lumi_at_x(pdf1, pdf2)

    ! O(as) contribution
    C1_x_pdf1 = (C1_plus_P .conv. pdf1)
    C1_x_pdf2 = (C1_plus_P .conv. pdf2)
    res = res + as2pi_mu*inclusive_lumi_at_x(C1_x_pdf1, pdf2)
    res = res + as2pi_mu*inclusive_lumi_at_x(pdf1, C1_x_pdf2)

    ! add virtual corrections
    res = res + cs%as2pi*HH1*inclusive_lumi_at_x(pdf1, pdf2)

    if(cs%prime) then
       as2pi_mu = cs%as2pi / (1-2*lambda)

       C2_x_pdf1 = (C2_plus_P .conv. pdf1)
       C2_x_pdf2 = (C2_plus_P .conv. pdf2)
       res = res + as2pi_mu**2*inclusive_lumi_at_x(C2_x_pdf1, pdf2)
       res = res + as2pi_mu**2*inclusive_lumi_at_x(pdf1, C2_x_pdf2)

       ! add cross products between the two coefficient functions
       ! (check that this is equivalent to eq. 19 of 1106.4652)
       res = res + as2pi_mu**2*inclusive_lumi_at_x(C1_x_pdf1, C1_x_pdf2)
       G1_x_pdf1 = (G1_matrix .conv. pdf1)
       G1_x_pdf2 = (G1_matrix .conv. pdf2)
       res = res + as2pi_mu**2*inclusive_lumi_at_x(G1_x_pdf1, G1_x_pdf2)

       ! add virtual corrections
       res = res + cs%as2pi**2*H(2)*inclusive_lumi_at_x(pdf1, pdf2)
       res = res + cs%as2pi*as2pi_mu*HH1*(inclusive_lumi_at_x(C1_x_pdf1, pdf2) &
            & + inclusive_lumi_at_x(pdf1, C1_x_pdf2))
    endif

  end function inclusive_lumi_NNLL_loc

  !=======================================================================================
  ! same as above, but for the NNNLL luminosity
  ! virtual corrections are part of the luminosity
  function inclusive_lumi_NNNLL_loc(muF, lambda) result(res)
    real(dp),                  intent(in) :: muF, lambda
    real(dp)                              :: res
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    real(dp) :: C1_x_pdf1(0:grid%ny,-6:7), C1_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: C2_x_pdf1(0:grid%ny,-6:7), C2_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: G1_x_pdf1(0:grid%ny,-6:7), G1_x_pdf2(0:grid%ny,-6:7)
    real(dp) :: as2pi_mu, HH1

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    if (cs%loop_mass .ne. 'none') then
       HH1 = Hm(1)
    else
       HH1 = H(1)
    end if

    call get_pdfs(muF, cs%collider, pdf1, pdf2)

    ! get the running coupling at scale muR = muR_default * exp(-L)
    as2pi_mu = cs%as2pi / (1-two*lambda) * (one &
         & - (twopi*cs%as2pi) / (1-two*lambda) * beta1/beta0 * log(one-two*lambda))

    ! get the luminosities w/o spurious O(as^3) terms
    ! O(as^0) contribution
    res = inclusive_lumi_at_x(pdf1, pdf2)

    ! O(as) contribution
    C1_x_pdf1 = (C1_plus_P .conv. pdf1)
    C1_x_pdf2 = (C1_plus_P .conv. pdf2)
    res = res + as2pi_mu*inclusive_lumi_at_x(C1_x_pdf1, pdf2)
    res = res + as2pi_mu*inclusive_lumi_at_x(pdf1, C1_x_pdf2)

    ! add virtual corrections
    res = res + cs%as2pi*HH1*inclusive_lumi_at_x(pdf1, pdf2)

    ! Now include O(as^2) corrections
    as2pi_mu = cs%as2pi / (1-2*lambda)
    C2_x_pdf1 = (C2_plus_P .conv. pdf1)
    C2_x_pdf2 = (C2_plus_P .conv. pdf2)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(C2_x_pdf1, pdf2)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(pdf1, C2_x_pdf2)

    ! add cross products between the two coefficient functions
    ! (check that this is equivalent to eq. 19 of 1106.4652)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(C1_x_pdf1, C1_x_pdf2)
    G1_x_pdf1 = (G1_matrix .conv. pdf1)
    G1_x_pdf2 = (G1_matrix .conv. pdf2)
    res = res + as2pi_mu**2*inclusive_lumi_at_x(G1_x_pdf1, G1_x_pdf2)

    ! add virtual corrections
    res = res + cs%as2pi**2*H(2)*inclusive_lumi_at_x(pdf1, pdf2)
    res = res + cs%as2pi*as2pi_mu*HH1*(inclusive_lumi_at_x(C1_x_pdf1, pdf2) &
      & + inclusive_lumi_at_x(pdf1, C1_x_pdf2))

  end function inclusive_lumi_NNNLL_loc

  !=======================================================================================
  ! The following two routines implement the additional pieces in our Eq 3.15 involving the convolution
  ! of the NLL luminosity with one and two splitting functions
    function inclusive_P0_lumi_NLL(muF, lambda) result(res)
    real(dp),                  intent(in) :: muF, lambda
    real(dp)             :: res
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    real(dp) :: as2pi_mu
    type(split_mat) :: C_plus_P
    integer  :: i
    type(pdf_rep) :: pdfrep
    real(dp) :: P0_conv_pdf1(0:grid%ny,-6:7), P0_conv_pdf2(0:grid%ny,-6:7)

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    ! get the pdf grids to compute convolutions with the coefficient functions
    P0_conv_pdf1=zero
    P0_conv_pdf2=zero
    call EvalPdfTable_Q(P0_PDFs, muF, P0_conv_pdf1)
    call EvalPdfTable_Q(P0_PDFs, muF, P0_conv_pdf2)
    call get_pdfs(muF, cs%collider, pdf1, pdf2)
    res = inclusive_lumi_at_x(pdf1, P0_conv_pdf2) + inclusive_lumi_at_x(P0_conv_pdf1, pdf2)
    ! insert coupling and beta function factor
    res = res * two*beta0/pi*cs%alphas_muR**2/(one-two*lambda)**2

  end function inclusive_P0_lumi_NLL

  function inclusive_P0_P0_lumi_NLL(muF, lambda) result(res)
    real(dp),                  intent(in) :: muF, lambda
    real(dp)             :: res
    !----------------------------------------------
    real(dp) :: pdf1(0:grid%ny,-6:7), pdf2(0:grid%ny,-6:7)
    real(dp) :: as2pi_mu
    type(split_mat) :: C_plus_P
    integer  :: i
    type(pdf_rep) :: pdfrep
    real(dp) :: P0_conv_pdf1(0:grid%ny,-6:7), P0_conv_pdf2(0:grid%ny,-6:7)
    real(dp) :: P0_P0_conv_pdf1(0:grid%ny,-6:7), P0_P0_conv_pdf2(0:grid%ny,-6:7)

    res = zero
    ! cutoff at the Landau pole
    if (lambda >= half) return

    ! get the pdf grids to compute convolutions with the coefficient functions
    P0_P0_conv_pdf1=zero
    P0_P0_conv_pdf2=zero
    call EvalPdfTable_Q(P0_P0_PDFs, muF, P0_P0_conv_pdf1)
    call EvalPdfTable_Q(P0_P0_PDFs, muF, P0_P0_conv_pdf2)
    P0_conv_pdf1=zero
    P0_conv_pdf2=zero
    call EvalPdfTable_Q(P0_PDFs, muF, P0_conv_pdf1)
    call EvalPdfTable_Q(P0_PDFs, muF, P0_conv_pdf2)
    call get_pdfs(muF, cs%collider, pdf1, pdf2)
    res = inclusive_lumi_at_x(pdf1, P0_P0_conv_pdf2) + inclusive_lumi_at_x(P0_P0_conv_pdf1, pdf2)
    res = res + two*inclusive_lumi_at_x(P0_conv_pdf1, P0_conv_pdf2)
    ! insert coupling and beta function factor
    res = res * cs%alphas_muR**2/pi**2/(one-two*lambda)**2

  end function inclusive_P0_P0_lumi_NLL

  !=======================================================================================
  !The following two functions are used to deallocate memory
  subroutine deallocate_grids
    implicit none

    call Delete (C1_matrix)
    call Delete (C2_matrix)
    call Delete (G1_matrix)
    call Delete (dglap_h)
    call Delete (PDFs)

  end subroutine deallocate_grids

  subroutine deallocate_C_matxs()
    implicit none

    call Delete(C1_PDFs)
    call Delete(C2_PDFs)
    call Delete(G1_PDFs)
    call Delete(P0_PDFs)
    call Delete(P0_P0_PDFs)
    call Delete (C1_plus_P)
    call Delete (C2_plus_P)
    call Delete (C1_x_P0_matrix)
    call Delete (P0_matrix)
    call Delete (P0_x_P0_matrix)
    call Delete (C1_matrix_PDFs)
    call Delete (C2_matrix_PDFs)
    call Delete (C1_P0_matrix_PDFs)

  end subroutine deallocate_C_matxs


end module pdfs_tools


!======================================================================
!======================================================================
!===== the following module contains routines which perform the  ======
!===== expansion of the luminosity factor up to order O(alphas^3) =====
!======================================================================
!======================================================================

module pdf_expansions
  use hoppet_v1; use rad_tools
  use pdfs_tools
  implicit none

  private :: Delete_PDFE
  public :: set_lumi_expansion_Born
  public :: set_pdf_expansion, set_pdf_expansion_at_x


  interface Delete
    module procedure Delete_PDFE, Delete_LumiE
  end interface

  !=======================================================================================
  !- type to hold the expansion of a PDF;
  type PDFExpansion
    ! pij is coefficient of (as2pi(muR))^i * t^j, where t = ln muF1^2/muF0^2 = -2L
    real(dp), pointer :: p(:,:)
    real(dp), pointer :: p11(:,:)
    real(dp), pointer :: p21(:,:)
    real(dp), pointer :: p22(:,:)
    real(dp), pointer :: p31(:,:)
    real(dp), pointer :: p32(:,:)
    real(dp), pointer :: p33(:,:)

    ! cij is part of coefficient of (as2pi(muR))^i * t^j, where t =
    ! ln muF1^2/muF0^2 = -2L that includes a factor of the coefficient
    ! function
    real(dp), pointer :: c10(:,:)
    real(dp), pointer :: c21(:,:)
    real(dp), pointer :: c20(:,:)
    real(dp), pointer :: c31(:,:)
    real(dp), pointer :: c32(:,:)

    real(dp), pointer :: g10(:,:)
    real(dp), pointer :: g21(:,:)
 end type PDFExpansion

 type PDFExpansion_at_x
    ! pij is coefficient of (as2pi(muR))^i * t^j, where t = ln muF1^2/muF0^2 = -2L
    real(dp) :: p(-6:7)
    real(dp) :: p11(-6:7)
    real(dp) :: p21(-6:7)
    real(dp) :: p22(-6:7)
    real(dp) :: p31(-6:7)
    real(dp) :: p32(-6:7)
    real(dp) :: p33(-6:7)

    ! cij is part of coefficient of (as2pi(muR))^i * t^j, where t =
    ! ln muF1^2/muF0^2 = -2L that includes a factor of the coefficient
    ! function
    real(dp) :: c10(-6:7)
    real(dp) :: c21(-6:7)
    real(dp) :: c20(-6:7)
    real(dp) :: c31(-6:7)
    real(dp) :: c32(-6:7)

    real(dp) :: g10(-6:7)
    real(dp) :: g21(-6:7)
 end type PDFExpansion_at_x

  !--------------------------
  type LumiExpansion
    real(dp), pointer :: l00(:)

    real(dp), pointer :: l10(:)
    real(dp), pointer :: l11(:)

    real(dp), pointer :: l20(:)
    real(dp), pointer :: dl21(:)
    real(dp), pointer :: l21(:)
    real(dp), pointer :: l22(:)

    real(dp), pointer :: l30(:)
    real(dp), pointer :: l31(:)
    real(dp), pointer :: l32(:)
    real(dp), pointer :: l33(:)

    real(dp), pointer :: p020(:)
    real(dp), pointer :: p0p020(:)
    real(dp), pointer :: p0p031(:)
  end type LumiExpansion

  !--------------------------
  type LumiExpansion_at_x
     real(dp) :: l00

     real(dp) :: l10
     real(dp) :: l11

     real(dp) :: l20
     real(dp) :: dl21
     real(dp) :: l21
     real(dp) :: l22

     real(dp) :: l30
     real(dp) :: l31
     real(dp) :: l32
     real(dp) :: l33

     real(dp) :: p020
     real(dp) :: p0p020
     real(dp) :: p0p031
  end type LumiExpansion_at_x



contains

  !---------------------------------------------------
  subroutine AllocPDFExpansion(pdfe)
    type(PDFExpansion), intent(inout) :: pdfe
    call AllocPDF(grid,pdfe%p  )
    call AllocPDF(grid,pdfe%p11)
    call AllocPDF(grid,pdfe%p21)
    call AllocPDF(grid,pdfe%p22)
    call AllocPDF(grid,pdfe%p31)
    call AllocPDF(grid,pdfe%p32)
    call AllocPDF(grid,pdfe%p33)

    call AllocPDF(grid,pdfe%c10)
    call AllocPDF(grid,pdfe%c21)
    call AllocPDF(grid,pdfe%c20)
    call AllocPDF(grid,pdfe%c31)
    call AllocPDF(grid,pdfe%c32)

    call AllocPDF(grid,pdfe%g10)
    call AllocPDF(grid,pdfe%g21)

  end subroutine AllocPDFExpansion

  !TODO: the following routine might not be needed; the pdfs are deleted
  !already in expansion.f90. We can keep this routine but in that case it's better if we call it.
  subroutine Delete_PDFE(pdfe)
    type(PDFExpansion), intent(inout) :: pdfe
    deallocate(pdfe%p  )
    deallocate(pdfe%p11)
    deallocate(pdfe%p21)
    deallocate(pdfe%p22)
    deallocate(pdfe%p31)
    deallocate(pdfe%p32)
    deallocate(pdfe%p33)

    deallocate(pdfe%c10)
    deallocate(pdfe%c21)
    deallocate(pdfe%c20)
    deallocate(pdfe%c31)
    deallocate(pdfe%c32)

    deallocate(pdfe%g10)
    deallocate(pdfe%g21)
  end subroutine Delete_PDFE

  !---------------------------------------------------
  subroutine AllocLumiExpansion(lumie)
    type(LumiExpansion), intent(inout) :: lumie
    call AllocGridQuant(grid,lumie%l00)
    call AllocGridQuant(grid,lumie%l10)
    call AllocGridQuant(grid,lumie%l11)
    call AllocGridQuant(grid,lumie%l20)
    call AllocGridQuant(grid,lumie%dl21)
    call AllocGridQuant(grid,lumie%l21)
    call AllocGridQuant(grid,lumie%l22)
    call AllocGridQuant(grid,lumie%l30)
    call AllocGridQuant(grid,lumie%l31)
    call AllocGridQuant(grid,lumie%l32)
    call AllocGridQuant(grid,lumie%l33)
    ! additional terms needed at N3LL
    call AllocGridQuant(grid,lumie%p020)
    call AllocGridQuant(grid,lumie%p0p020)
    call AllocGridQuant(grid,lumie%p0p031)
  end subroutine AllocLumiExpansion

  !TODO: the following routine might not be needed; the luminosities are deleted
  !already in expansion.f90. We can keep this routine but in that case it's better if we call it.
  subroutine Delete_LumiE(lumie)
    type(LumiExpansion), intent(inout) :: lumie
    deallocate(lumie%l00)
    deallocate(lumie%l10)
    deallocate(lumie%l11)
    deallocate(lumie%l20)
    deallocate(lumie%dl21)
    deallocate(lumie%l21)
    deallocate(lumie%l22)
    deallocate(lumie%l30)
    deallocate(lumie%l31)
    deallocate(lumie%l32)
    deallocate(lumie%l33)
    deallocate(lumie%p020)
    deallocate(lumie%p0p020)
    deallocate(lumie%p0p031)
  end subroutine Delete_LumiE



  !=======================================================================================
  ! given an input pdf that has already been set in pdfe%p (assuming
  ! the whole of pdfe has been allocated),
  subroutine set_pdf_expansion(pdfe,u)
    type(PDFExpansion), intent(inout) :: pdfe
    real(dp),           intent(in)    :: u ! in the final expansion u = ln(muF^2/muR^2)
    !----------------------------------
    real(dp) :: P0_pdf(size(pdfe%p,1),size(pdfe%p,2))
    real(dp) :: P0_P0_pdf(size(pdfe%p,1),size(pdfe%p,2)), P1_pdf(size(pdfe%p,1),size(pdfe%p,2))
    real(dp) :: P2_pdf(size(pdfe%p,1),size(pdfe%p,2))
    real(dp) :: P1_P0_pdf(size(pdfe%p,1),size(pdfe%p,2)), P0_P1_pdf(size(pdfe%p,1),size(pdfe%p,2))
    real(dp) :: P0_P0_P0_pdf(size(pdfe%p,1),size(pdfe%p,2))

    type(split_mat) :: C1, C2, G1
    type(split_mat) :: P0_x_P0, C1_x_P0

    call AllocSplitMat(grid, C1, nf_int)
    call AllocSplitMat(grid, C2, nf_int)
    call AllocSplitMat(grid, G1, nf_int)

    call AllocSplitMat(grid, P0_x_P0, nf_int)
    call AllocSplitMat(grid, C1_x_P0, nf_int)

    !--- set up the convolutions of splitting fn * pdf
    P0_pdf = dglap_h%P_LO * pdfe%p
    P0_P0_pdf = dglap_h%P_LO * P0_pdf
    P0_P0_P0_pdf = dglap_h%P_LO * P0_P0_pdf

    P1_pdf = dglap_h%P_NLO * pdfe%p
    P1_P0_pdf = dglap_h%P_NLO * P0_pdf
    P0_P1_pdf = dglap_h%P_LO * P1_pdf

    P2_pdf = dglap_h%P_NNLO * pdfe%p

    !--- now work out the various expansions
    pdfe%p11 = P0_pdf

    pdfe%p22 = half*P0_P0_pdf - pi*beta0 * P0_pdf
    pdfe%p21 = P1_pdf - twopi*u*beta0 * P0_pdf

    pdfe%p33 = (P0_P0_P0_pdf - (6*Pi*beta0)*P0_P0_pdf + (8*Pi**2*beta0**2)*P0_pdf)/6._dp
    pdfe%p32 = (P0_P1_pdf + P1_P0_pdf)/2._dp - (2*Pi*u*beta0)*P0_P0_pdf -  &
         &    (2*Pi*beta0)*P1_pdf + (4*Pi**2*u*beta0**2- 2*Pi**2*beta1)*P0_pdf
    pdfe%p31 = P2_pdf - (4*Pi*u*beta0)*P1_pdf + (4*Pi**2*u*(u*beta0**2-beta1))*P0_pdf

    ! these pieces should add as2pi(muF1/muF0 * muR) * C1 * pdf_evolved
    ! as2pi(muF1/muF0 * muR) = as2pi(muR) - b0*t*as2pi^2
    call InitSplitMat(C1, C1_matrix)
    call AddWithCoeff(C1, dglap_h%P_LO, -cs%ln_muF2_M2+cs%ln_Q2_M2)

    pdfe%c10 = C1*pdfe%p
    pdfe%c21 = C1*pdfe%p11 - twopi*beta0*pdfe%c10
    pdfe%c31 = C1*pdfe%p21
    pdfe%c32 = C1*pdfe%p22 - twopi*beta0*(C1*pdfe%p11) + twopi**2*beta0**2*pdfe%c10


    ! include N3LL pieces if required
    pdfe%c20 = zero; pdfe%g10 = zero; pdfe%g21 = zero
    if ( (cs%res_order == order_NNNLL) .or. ( (cs%res_order == order_NNLL) .and. (cs%prime)) ) then
       pdfe%c31 = pdfe%c31 - twopi**2*beta1*pdfe%c10  ! this last piece appears if the two-loop running were used in the coefficient function instead
       ! O(as^2) coefficient functions
       call InitSplitMat(C2, C2_matrix)
       ! add resummation scale dependence
       call AddWithCoeff(C2, dglap_h%P_LO, pi*beta0*(-cs%ln_muF2_M2+cs%ln_Q2_M2)**2 &
            & - two*pi*beta0*(-cs%ln_muF2_M2+cs%ln_Q2_M2)*cs%ln_Q2_muR2)
       call AddWithCoeff(C2, dglap_h%P_NLO, -cs%ln_muF2_M2+cs%ln_Q2_M2)
       call SetToConvolution(P0_x_P0, dglap_h%P_LO, dglap_h%P_LO)
       call AddWithCoeff(C2, P0_x_P0, half*(-cs%ln_muF2_M2+cs%ln_Q2_M2)**2)
       call SetToConvolution(C1_x_P0, C1_matrix, dglap_h%P_LO)
       call AddWithCoeff(C2, C1_x_P0, -cs%ln_muF2_M2+cs%ln_Q2_M2)
       call AddWithCoeff(C2, C1_matrix, -two*pi*beta0*cs%ln_Q2_muR2)

       pdfe%c20 = C2*pdfe%p
       pdfe%c31 = pdfe%c31 + C2*P0_pdf - four*pi*beta0*pdfe%c20

       ! define contributions from the G1 coefficient functions necessary for a O(as^3) expansion
       ! (notice they don't multiply the Ci functions at any order)
       call InitSplitMat(G1, G1_matrix)
       pdfe%g10 = G1*pdfe%p
       pdfe%g21 = G1*P0_pdf - two*pi*beta0*pdfe%g10
    end if

    call Delete(C1)
    call Delete(C2)
    call Delete(G1)
    call Delete(P0_x_P0)
    call Delete(C1_x_P0)

  end subroutine set_pdf_expansion

  !=======================================================================================
  ! The following routine returns the luminosity expansion up to O(as^3)
  ! for a given colour singlet process
  ! pdfe1 and pdfe2 are the two pdf expansions evaluated at x1 and x2 respectively
  ! in the expansion routine
  subroutine set_lumi_expansion_Born(pdf1e, pdf2e, msqB, msqV1, msqV2, lumie, order)
    type(PDFExpansion_at_x),        intent(in)  :: pdf1e, pdf2e
    real(dp), intent(in) :: msqB(-nf_int:nf_int,-nf_int:nf_int), msqV1(-nf_int:nf_int,-nf_int:nf_int), msqV2(-nf_int:nf_int,-nf_int:nf_int)
    type(LumiExpansion_at_x),     intent(out) :: lumie
    integer,                      intent(in)  :: order ! resummation order


    lumie%l00 = lumi_Born(pdf1e%p, pdf2e%p, msqB)
    lumie%l11 = lumi_Born(pdf1e%p11, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%p11,msqB)
    lumie%l21 = lumi_Born(pdf1e%p21, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%p21,msqB)
    lumie%l22 = lumi_Born(pdf1e%p11, pdf2e%p11,msqB) + lumi_Born(pdf1e%p22, pdf2e%p,msqB) +&
         & lumi_Born(pdf1e%p, pdf2e%p22,msqB)
    lumie%l31 = lumi_Born(pdf1e%p31, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%p31,msqB)
    lumie%l32 = lumi_Born(pdf1e%p32, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%p32,msqB) + &
         & lumi_Born(pdf1e%p21, pdf2e%p11,msqB) + lumi_Born(pdf1e%p11, pdf2e%p21,msqB)
    lumie%l33 = lumi_Born(pdf1e%p33, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%p33,msqB) + &
         & lumi_Born(pdf1e%p22, pdf2e%p11,msqB) + lumi_Born(pdf1e%p11, pdf2e%p22,msqB)


    lumie%l10  = zero
    lumie%l20  = zero
    lumie%dl21 = zero ! this was used in ptjet resummation, useless here, but we want to keep it for consistency
    lumie%l30  = zero

    ! initialise N3LL terms
    lumie%p020   = zero
    lumie%p0p020 = zero
    lumie%p0p031 = zero

    ! coefficient functions are included only at NNLL for now
    ! add the coefficient functions also if prime counting is active
    if ( (order > 1) .or. (order == 1 .and. cs%prime) ) then
       lumie%l10  = lumie%l10  + lumi_Born(pdf1e%c10, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%c10,msqB)
       lumie%l21  = lumie%l21  + lumi_Born(pdf1e%c21, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%c21,msqB)&
            &                  + lumi_Born(pdf1e%c10, pdf2e%p11,msqB) + lumi_Born(pdf1e%p11, pdf2e%c10,msqB)
       lumie%l31  = lumie%l31  + lumi_Born(pdf1e%c31, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%c31,msqB)&
            &                  + lumi_Born(pdf1e%c10, pdf2e%p21,msqB) + lumi_Born(pdf1e%p21, pdf2e%c10,msqB)

       lumie%l32  = lumie%l32  + lumi_Born(pdf1e%c32, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%c32,msqB)&
            &                  + lumi_Born(pdf1e%c21, pdf2e%p11,msqB) + lumi_Born(pdf1e%p11, pdf2e%c21,msqB)&
            &                  + lumi_Born(pdf1e%c10, pdf2e%p22,msqB) + lumi_Born(pdf1e%p22, pdf2e%c10,msqB)


       ! add virtual corrections O(as)
       lumie%l10  = lumie%l10  + lumi_Born(pdf1e%p, pdf2e%p, msqV1)
       !lumie%l10  = lumie%l10  + HH1*lumie%l00
       lumie%l31  = lumie%l31  + lumi_Born(pdf1e%p21, pdf2e%p,msqV1) + lumi_Born(pdf1e%p, pdf2e%p21,msqV1) &
            & + lumi_Born(pdf1e%c21, pdf2e%p,msqV1) + lumi_Born(pdf1e%p, pdf2e%c21,msqV1) &
            & + lumi_Born(pdf1e%c10, pdf2e%p11,msqV1) + lumi_Born(pdf1e%p11, pdf2e%c10,msqV1)
       !lumie%l31  = lumie%l31  + HH1*lumie%l21
       lumie%l21  = lumie%l21  + lumi_Born(pdf1e%p11, pdf2e%p,msqV1) + lumi_Born(pdf1e%p, pdf2e%p11,msqV1)
       !lumie%l21  = lumie%l21  + HH1*lumie%l11
       lumie%l32  = lumie%l32  + lumi_Born(pdf1e%p11, pdf2e%p11,msqV1) + lumi_Born(pdf1e%p22, pdf2e%p,msqV1) &
            & + lumi_Born(pdf1e%p, pdf2e%p22,msqV1)
       !lumie%l32  = lumie%l32  + HH1*lumie%l22
    end if

    ! add the coefficient functions also if prime counting is active
    if ( (order > 2) .or. (order == 2 .and. cs%prime) ) then
       ! add N3LL piece
       ! include contributions from C2 and G1 related terms
       lumie%l20 = lumi_Born(pdf1e%c10, pdf2e%c10,msqB) &
            & + lumi_Born(pdf1e%c20, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%c20,msqB) &
            & + lumi_Born(pdf1e%g10, pdf2e%g10,msqB)
       lumie%l31 = lumie%l31 + lumi_Born(pdf1e%c21, pdf2e%c10,msqB) + lumi_Born(pdf1e%c10, pdf2e%c21,msqB) &
            & + lumi_Born(pdf1e%c20, pdf2e%p11,msqB) + lumi_Born(pdf1e%p11, pdf2e%c20,msqB) &
            & + lumi_Born(pdf1e%g10, pdf2e%g21,msqB) + lumi_Born(pdf1e%g21, pdf2e%g10,msqB)

       ! Add virtual corrections at O(as^2)
       lumie%l20 = lumie%l20 + lumi_Born(pdf1e%p, pdf2e%p, msqV2) &
            & + (lumi_Born(pdf1e%c10, pdf2e%p,msqV1) + lumi_Born(pdf1e%p, pdf2e%c10,msqV1))
       lumie%l31 = lumie%l31 + (lumi_Born(pdf1e%p11, pdf2e%p,msqV2) + lumi_Born(pdf1e%p, pdf2e%p11,msqV2))

       ! now build the extra terms appearing at N3LL
       !! P0_lumi_NLL = two*beta0/pi*cs%alphas_muR**2/(one-two*lambda)**2*P0xLNLL
       !! P0_lumi_NLL = lumie%PO20 (as/twopi)^2
       !! lumie%POij (as(muR)/twopi)^i (-2 * L)^j
       !! P0_P0_lumi_NLL = cs%alphas_muR**2/pi**2/(one-two*lambda)**2*P0xP0xLNLL
       !! P0_P0_lumi_NLL=lumie%POPO20 (as/twopi) + lumie%POPO31 (as/twopi)^2 (-2 L)
       !! where the coefficients are given by:
       lumie%p020   = 8._dp*pi*beta0*(lumi_Born(pdf1e%p11, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%p11,msqB))
       lumie%p0p020 = 8._dp*(lumi_Born(pdf1e%p11, pdf2e%p11,msqB) + lumi_Born(pdf1e%p22, pdf2e%p,msqB) +&
            & lumi_Born(pdf1e%p, pdf2e%p22,msqB)) + lumie%p020
       lumie%p0p031 = 24._dp*(lumi_Born(pdf1e%p33, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%p33,msqB) + &
            & lumi_Born(pdf1e%p22, pdf2e%p11,msqB) + lumi_Born(pdf1e%p11, pdf2e%p22,msqB)) &
            & + 16._dp*pi*beta0*(lumi_Born(pdf1e%p11, pdf2e%p11,msqB) + lumi_Born(pdf1e%p22, pdf2e%p,msqB) +&
            & lumi_Born(pdf1e%p, pdf2e%p22,msqB)) &
            & - 16._dp*pi**2*beta0**2*(lumi_Born(pdf1e%p11, pdf2e%p,msqB) + lumi_Born(pdf1e%p, pdf2e%p11,msqB))
    end if

  end subroutine set_lumi_expansion_Born


  subroutine set_pdf_expansion_at_x(pdf1eatx, pdf2eatx, pdf1e, pdf2e, x1, x2)
    type(PDFExpansion_at_x),    intent(out) :: pdf1eatx, pdf2eatx
    type(PDFExpansion),       intent(in) :: pdf1e, pdf2e
    real(dp), intent(in) :: x1, x2

    pdf1eatx%p = pdf1e%p .atx. (x1 .with. grid)
    pdf2eatx%p = pdf2e%p .atx. (x2 .with. grid)

    pdf1eatx%p11 = pdf1e%p11 .atx. (x1 .with. grid)
    pdf2eatx%p11 = pdf2e%p11 .atx. (x2 .with. grid)

    pdf1eatx%p21 = pdf1e%p21 .atx. (x1 .with. grid)
    pdf2eatx%p21 = pdf2e%p21 .atx. (x2 .with. grid)

    pdf1eatx%p22 = pdf1e%p22 .atx. (x1 .with. grid)
    pdf2eatx%p22 = pdf2e%p22 .atx. (x2 .with. grid)

    pdf1eatx%p31 = pdf1e%p31 .atx. (x1 .with. grid)
    pdf2eatx%p31 = pdf2e%p31 .atx. (x2 .with. grid)

    pdf1eatx%p32 = pdf1e%p32 .atx. (x1 .with. grid)
    pdf2eatx%p32 = pdf2e%p32 .atx. (x2 .with. grid)

    pdf1eatx%p33 = pdf1e%p33 .atx. (x1 .with. grid)
    pdf2eatx%p33 = pdf2e%p33 .atx. (x2 .with. grid)

    pdf1eatx%c10 = pdf1e%c10 .atx. (x1 .with. grid)
    pdf2eatx%c10 = pdf2e%c10 .atx. (x2 .with. grid)

    pdf1eatx%c21 = pdf1e%c21 .atx. (x1 .with. grid)
    pdf2eatx%c21 = pdf2e%c21 .atx. (x2 .with. grid)

    pdf1eatx%c20 = pdf1e%c20 .atx. (x1 .with. grid)
    pdf2eatx%c20 = pdf2e%c20 .atx. (x2 .with. grid)

    pdf1eatx%c31 = pdf1e%c31 .atx. (x1 .with. grid)
    pdf2eatx%c31 = pdf2e%c31 .atx. (x2 .with. grid)

    pdf1eatx%c32 = pdf1e%c32 .atx. (x1 .with. grid)
    pdf2eatx%c32 = pdf2e%c32 .atx. (x2 .with. grid)

    pdf1eatx%g10 = pdf1e%g10 .atx. (x1 .with. grid)
    pdf2eatx%g10 = pdf2e%g10 .atx. (x2 .with. grid)

    pdf1eatx%g21 = pdf1e%g21 .atx. (x1 .with. grid)
    pdf2eatx%g21 = pdf2e%g21 .atx. (x2 .with. grid)

  end subroutine set_pdf_expansion_at_x

  ! The following routine returns the luminosity expansion for ptB resummation up to O(as^3)
  ! virtual corrections are treated explicitely in the expansion
  subroutine inclusive_set_lumi_expansion_ptB(pdf1e, pdf2e, lumie, order)
    type(PDFExpansion),           intent(in)  :: pdf1e, pdf2e
    type(LumiExpansion),          intent(out) :: lumie
    integer,                      intent(in)  :: order ! resummation order
    real(dp)                                  :: HH1

    lumie%l00 = inclusive_lumi_all_x(pdf1e%p, pdf2e%p)
    lumie%l11 = inclusive_lumi_all_x(pdf1e%p11, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%p11)
    lumie%l21 = inclusive_lumi_all_x(pdf1e%p21, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%p21)
    lumie%l22 = inclusive_lumi_all_x(pdf1e%p11, pdf2e%p11) + inclusive_lumi_all_x(pdf1e%p22, pdf2e%p) +&
            & inclusive_lumi_all_x(pdf1e%p, pdf2e%p22)
    lumie%l31 = inclusive_lumi_all_x(pdf1e%p31, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%p31)
    lumie%l32 = inclusive_lumi_all_x(pdf1e%p32, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%p32) + &
            & inclusive_lumi_all_x(pdf1e%p21, pdf2e%p11) + inclusive_lumi_all_x(pdf1e%p11, pdf2e%p21)
    lumie%l33 = inclusive_lumi_all_x(pdf1e%p33, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%p33) + &
            & inclusive_lumi_all_x(pdf1e%p22, pdf2e%p11) + inclusive_lumi_all_x(pdf1e%p11, pdf2e%p22)


    lumie%l10  = zero
    lumie%l20  = zero
    lumie%dl21 = zero ! this was used in ptjet resummation, useless here, but we want to keep it for consistency
    lumie%l30  = zero

    ! coefficient functions are included only at NNLL for now
    if ( (order > 1) .or. (order == 1 .and. cs%prime) ) then
       lumie%l10  = lumie%l10  + inclusive_lumi_all_x(pdf1e%c10, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%c10)
       lumie%l21  = lumie%l21  + inclusive_lumi_all_x(pdf1e%c21, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%c21)&
            &                  + inclusive_lumi_all_x(pdf1e%c10, pdf2e%p11) + inclusive_lumi_all_x(pdf1e%p11, pdf2e%c10)
       lumie%l31  = lumie%l31  + inclusive_lumi_all_x(pdf1e%c31, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%c31)&
            &                  + inclusive_lumi_all_x(pdf1e%c10, pdf2e%p21) + inclusive_lumi_all_x(pdf1e%p21, pdf2e%c10)

       lumie%l32  = lumie%l32  + inclusive_lumi_all_x(pdf1e%c32, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%c32)&
            &                  + inclusive_lumi_all_x(pdf1e%c21, pdf2e%p11) + inclusive_lumi_all_x(pdf1e%p11, pdf2e%c21)&
            &                  + inclusive_lumi_all_x(pdf1e%c10, pdf2e%p22) + inclusive_lumi_all_x(pdf1e%p22, pdf2e%c10)


       ! add virtual corrections O(as)
       if (cs%loop_mass .ne. 'none') then
          HH1 = Hm(1)
       else
          HH1 = H(1)
       end if
       lumie%l10  = lumie%l10  + HH1*lumie%l00
       lumie%l31  = lumie%l31  + HH1*lumie%l21
       lumie%l21  = lumie%l21  + HH1*lumie%l11
       lumie%l32  = lumie%l32  + HH1*lumie%l22

    end if

    if (order > 2 .or. (order == 2 .and. cs%prime) ) then
      ! include contributions from C2 and G1 related terms
      lumie%l20 = inclusive_lumi_all_x(pdf1e%c10, pdf2e%c10) &
               & + inclusive_lumi_all_x(pdf1e%c20, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%c20) &
               & + inclusive_lumi_all_x(pdf1e%g10, pdf2e%g10)
      lumie%l31 = lumie%l31 + inclusive_lumi_all_x(pdf1e%c21, pdf2e%c10) + inclusive_lumi_all_x(pdf1e%c10, pdf2e%c21) &
               & + inclusive_lumi_all_x(pdf1e%c20, pdf2e%p11) + inclusive_lumi_all_x(pdf1e%p11, pdf2e%c20) &
               & + inclusive_lumi_all_x(pdf1e%g10, pdf2e%g21) + inclusive_lumi_all_x(pdf1e%g21, pdf2e%g10)

       if (cs%loop_mass .ne. 'none') then
          HH1 = Hm(1)
       else
          HH1 = H(1)
       end if

      lumie%l20 = lumie%l20 + H(2)*lumie%l00 &
               & + HH1*(inclusive_lumi_all_x(pdf1e%c10, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%c10))
      lumie%l31 = lumie%l31 + H(2)*(inclusive_lumi_all_x(pdf1e%p11, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%p11))


      ! now build the extra terms appearing at N3LL
      !! P0_lumi_NLL = two*beta0/pi*cs%alphas_muR**2/(one-two*lambda)**2*P0xLNLL
      !! P0_lumi_NLL = lumie%PO20 (as/twopi)^2
      !! lumie%POij (as(muR)/twopi)^i (-2 * L)^j
      !! P0_P0_lumi_NLL = cs%alphas_muR**2/pi**2/(one-two*lambda)**2*P0xP0xLNLL
      !! P0_P0_lumi_NLL=lumie%POPO20 (as/twopi) + lumie%POPO31 (as/twopi)^2 (-2 L)
      !! where the coefficients are given by:
      !! TODO to be checked
      lumie%p020   = 8._dp*pi*beta0*(inclusive_lumi_all_x(pdf1e%p11, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%p11))
      lumie%p0p020 = 8._dp*(inclusive_lumi_all_x(pdf1e%p11, pdf2e%p11) + inclusive_lumi_all_x(pdf1e%p22, pdf2e%p) +&
            & inclusive_lumi_all_x(pdf1e%p, pdf2e%p22)) + lumie%p020
      lumie%p0p031 = 24._dp*(inclusive_lumi_all_x(pdf1e%p33, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%p33) + &
            & inclusive_lumi_all_x(pdf1e%p22, pdf2e%p11) + inclusive_lumi_all_x(pdf1e%p11, pdf2e%p22)) &
            & + 16._dp*pi*beta0*(inclusive_lumi_all_x(pdf1e%p11, pdf2e%p11) + inclusive_lumi_all_x(pdf1e%p22, pdf2e%p) +&
            & inclusive_lumi_all_x(pdf1e%p, pdf2e%p22)) &
            & - 16._dp*pi**2*beta0**2*(inclusive_lumi_all_x(pdf1e%p11, pdf2e%p) + inclusive_lumi_all_x(pdf1e%p, pdf2e%p11))
     end if

  end subroutine inclusive_set_lumi_expansion_ptB

  !=======================================================================================
  !--- subroutine that evaluates the luminosity at a given x (threshold)
  subroutine inclusive_set_lumi_expansion_at_x(lumi, lumie)
    type(LumiExpansion_at_x),    intent(out) :: lumi
    type(LumiExpansion),       intent(inout) :: lumie

    lumi%l00  = lumie%l00 .atx. (cs%M2_rts2 .with. grid)
    lumi%l10  = lumie%l10 .atx. (cs%M2_rts2 .with. grid)
    lumi%l11  = lumie%l11 .atx. (cs%M2_rts2 .with. grid)
    lumi%l20  = lumie%l20 .atx. (cs%M2_rts2 .with. grid)
    lumi%dl21 = lumie%dl21 .atx. (cs%M2_rts2 .with. grid)
    lumi%l21  = lumie%l21 .atx. (cs%M2_rts2 .with. grid)
    lumi%l22  = lumie%l22 .atx. (cs%M2_rts2 .with. grid)
    lumi%l30  = lumie%l30 .atx. (cs%M2_rts2 .with. grid)
    lumi%l31  = lumie%l31 .atx. (cs%M2_rts2 .with. grid)
    lumi%l32  = lumie%l32 .atx. (cs%M2_rts2 .with. grid)
    lumi%l33  = lumie%l33 .atx. (cs%M2_rts2 .with. grid)

    lumi%p020   = lumie%p020 .atx. (cs%M2_rts2 .with. grid)
    lumi%p0p020 = lumie%p0p020 .atx. (cs%M2_rts2 .with. grid)
    lumi%p0p031 = lumie%p0p031 .atx. (cs%M2_rts2 .with. grid)

  end subroutine inclusive_set_lumi_expansion_at_x

end module pdf_expansions
